#ifndef _AwlP_h
#define _AwlP_h

/* $Header: /usr/src/local/awl/RCS/AwlP.h,v 2.3 90/04/19 20:04:05 jkh Exp $ */

/*
 *
 *                   Copyright 1989, 1990
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * This file contains the Awl Widget private declarations and data.
 *
 * $Log:	AwlP.h,v $
 * Revision 2.3  90/04/19  20:04:05  jkh
 * Alpha checkin.
 * 
 * Revision 2.2  90/03/30  22:01:17  jkh
 * Added alu_table_size
 * 
 * Revision 2.1  90/03/30  01:06:51  jkh
 * Added forward ref for do_optional_jump().
 * 
 * Revision 2.0  90/03/26  01:42:07  jkh
 * pre-beta check-in
 * 
 */

#include <X11/IntrinsicP.h>
#include <X11/ShellP.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <setjmp.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <signal.h>
#include <varargs.h>
#ifdef PCS
#include <values.h>
#endif
#include "regex.h"

/* #include <X11/Awl.h> */
#include "Awl.h"

#include "macros.h"
#include "hash.h"
#include "ltypes.h"
#include "patchlevel.h"

typedef struct {
     int nichts;
} AwlClassPart;

typedef struct _AwlClassRec {
     CoreClassPart		core_class;
     CompositeClassPart		composite_class;
     ShellClassPart		shell_class;
     WMShellClassPart		wm_shell_class;
     VendorShellClassPart	vendor_shell_class;
     TopLevelShellClassPart	top_level_shell_class;
     AwlClassPart		awl_class;
} AwlClassRec;

extern AwlClassRec awlClassRec;

typedef struct {
     /* resources */
     struct {
	  Boolean debug;		/* debug parser */
	  Boolean lineinfo;		/* keep line number information */
	  Boolean nowarn;		/* don't output warnings */
	  Boolean showasm;		/* show assembly as generated */
	  Boolean trace;		/* show insn trace */
	  Boolean version;		/* print version information */
	  String cpp_args;		/* arguments to "cpp" command */
	  String cpp_command;		/* command name for "cpp" */
	  String file;			/* filename to parse */
	  String *argv;			/* optional starting arg vector */
	  String startfunc;		/* function to execute on startup */
	  AwlExternList externlist;	/* application externals */
     } res;
     /* private state */
     struct {
	  Table top;			/* the global symbol table */
	  jmp_buf bomb_out;		/* where to jump to on error */
	  CTable ctable;		/* global color table (optional) */
     } sysinfo;
     struct {
	  FILE *in, *out, *err;		/* widget stdin/stdout/stderr */
	  String sep;			/* what's magic for newline */
	  int radix;			/* i/o radix */
     } io;
     struct {
	  LTabEnt *ltab;		/* the line number table */
	  int ltabsize, ltabcurr;	/* max size, last entry */
	  Breakpoint bkpts;		/* breakpoint list */
     } debuginfo;
     struct {
	  Datum *text;			/* the text segment */
	  Value *stack;			/* the value stack */
	  int etext;			/* the end of the "text" segment */
	  int estack;			/* the end of the value stack */
     } seginfo;
     struct {
	  int sp;			/* the stack pointer */
	  int fp;			/* the frame pointer */
	  int pc;			/* the program counter */
	  int oldsp;			/* the saved SP */
	  int nparms;			/* number of parms on stack */
     } segptrs;
     struct {
	  int lastpc;			/* current compilation position */
	  int lastinsn;			/* for optimization */
	  Layout active;		/* intermediate widget layout */
	  Layout current;		/* for layout declarations */
     } compile;
} AwlPart;

typedef struct _AwlRec {
     CorePart		core;
     CompositePart	composite;
     ShellPart		shell;
     WMShellPart     wm;
     VendorShellPart vendor;
     TopLevelShellPart topLevel;
     AwlPart		awl;
} AwlRec;

/* system routines */
#ifdef __GNUC__
#define alloca __builtin_alloca
#else
Import char *alloca();
#endif
Import char *XtMalloc(), *bcopy();
Import int abs();
Import void bzero(), exit();

/* mapping routines from various gperf parsers */
Import WidgetClass	map_wname_to_wclass();
Import String		map_wclass_to_wname();
Import long		map_wclass_to_flags();
Import int		map_resword_to_token();
Import String		map_token_to_resword();
Import String		map_rname();

/* utility routines */
Import void backslash_eliminate(), append_string(), list_free();
Import Layout new_wl();
Import Generic convert_external();
Import Resource new_res();
Import Breakpoint new_bkpt();
Import AllocObj new_aobj();
Import AllocObj add_alloc(), sub_alloc(), mult_alloc(), div_alloc();
Import int sappend_string(), atonum(), strcomp(), stricomp(), list_len();
Import String numtoa(), ftoa(), s_strsed(), strsed(), skip_whitespace();
Import String datum_string_value(), value_string_value(), string_multiply();
Import String list_to_string();
Import String *list_from_string();
Import String *list_from_alist();
Import String *list_dup();
Import String format_time(), expand_stat();
Import Value coerce();
Import Process p_popen();
Import int p_pclose();

/* parsing routines */
Import int yyerror();
Import void yywarn();
Import int yylex(), yyparse();

/* disassembly/code info routines */
Import int op_nargs(), named_type();
Import void disasm();
Import LTabEnt *ltab_for_pc();
Import String op_name(), alu_name(), reg_name(), type_name(),
     class_name(), describe_symbol();

/* Symbol routines */
Import void		symbol_add();
Import void		symbol_remove();
Import void		symbol_def_constant();
Import void		symbol_def_variable();
Import void		symbol_set_locals();
Import Symbol		symbol_new();
Import Symbol		symbol_add_function();
Import Symbol		symbol_find();
Import Symbol		symbol_alist_find();
Import Symbol		symbol_def_function();
Import String		symbol_printname();
Import WidgetClass	lookup_client_wclass();

/* Widget creation and manipulation routines */
Import void wset_arg_by_value(), create_widget();
Import void layout_create(), layout_destroy(), layout_realize();
Import void layout_manage(), layout_unmanage();
Import void layout_map(), layout_unmap();
Import void layout_popup(), layout_popdown();
Import Widget (*wcreate_rtn())();
Import WidgetClass layout_wclass();
Import Widget layout_id();

/* Compiler/Evaluator main functions */
Import Value pop(), tos(), Eval(), do_load(), do_register_load();
Import void do_init(), do_emit(), do_alu();
Import void do_push(), do_store();
Import void do_discard(), do_optional_jump(), do_link(), do_rts();
Import void push();

/* Evaluator utilities */
Import Value *do_dot_lookup(), do_convert(), get_arg();
Import Resource do_res_lookup();
Import Symbol do_new_label();
Import void adjust(), do_free();
Import void exec_error(), exec_warn(), debug();
Import void do_callback_user(), do_timer_user();
Import void do_action_user(), do_signal_user();
Import void do_workproc_user();
Import int parseFile();

/* Interesting global variables */
Import Datum TrueDatum, FalseDatum;
Import Datum PopDupValue, PopDupTarget, PopDupNone, PopDiscard;
Import InsnTable opcode_table[], alu_opcodes[];
Import int opcode_table_size, alu_table_size;

/* from yylex() */
Import char *yy_file;
Import int yy_line;

#endif /* _AwlP_h */
