%{
#ifndef lint
static char *rcsid = "$Header: /usr/src/local/awl/RCS/awl.lex,v 2.4 90/04/19 20:04:21 jkh Exp $";
#endif

/*
 *
 *                     Copyright 1989
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * Here's awl's main lexer.
 *
 * $Log:	awl.lex,v $
 * Revision 2.4  90/04/19  20:04:21  jkh
 * Alpha checkin.
 * 
 * Revision 2.3  90/04/13  16:32:06  jkh
 * Added kludge to differentiate case (and default:) labels
 * and ordinary labels properly.
 * 
 * Revision 2.2  90/03/29  18:27:58  jkh
 * Fixed labels. Added 3 new operators.
 * 
 * Revision 2.0  90/03/26  01:42:44  jkh
 * pre-beta check-in
 * 
 */

#include "AwlP.h"
#include "y.tab.h"

/*
 * The next two vars are for error messages. I know about yylineno
 * but found out that flex doesn't support it, so I use my own
 * counter. This way you can use flex or lex [BTW, Vern, yylineno
 * *is* documented for SYSV lex]
 */
Export char *yy_file = NULL;
Export int yy_line;

Import Widget _currW;

/*
 * We need this cruft to deal with the non-portable way flex deals
 * with EOF (non-portable == pathologically incompatable with lex).
 */
#ifdef FLEX_SCANNER
Export lexer_reset()
{
     Import int yy_init;

     yy_init = 1;
}
#else
Export lexer_reset()
{}
#endif

%}
decdigit	[0-9]
hexdigit	[0-9A-Fa-f]
octdigit	[0-7]
uniqueid	\$\?
id		[a-zA-Z_$][a-zA-Z_$0-9]*
white		[ \t]+
any		[^\n]
dot		\.
float		{decdigit}*{dot}{decdigit}*
nl		[\n]
dq-string	\"([^"]|\\["])*\"
sq-string	\'[^\'\n]*\'
string		{dq-string}
hexnum		0[xX]{hexdigit}+
octnum		0{octdigit}+
number		{decdigit}+

%%

{nl}				yy_line++;
{white}					 ;

"=="				return(EE);
"<="				return(LE);
">="				return(GE);
"!="				return(NE);
"!!"				return(NUKE);
"||"				return(LOR);
"&&"				return(LAND);
"+="				return(PE);
"-="				return(ME);
"*="				return(TE);
"/="				return(DE);
"%="				return(RE);
"^="				return(XE);
"|="				return(OE);
"&="				return(AE);
"~="				return(CE);
"<<="				return(SLE);
">>="				return(SRE);
"++"				return(PP);
"--"				return(MM);
"<<"				return(SL);
">>"				return(SR);
".."				return(RANGE);
"!!"				return(NUKE);
"??"				return(MBR);
"<>"				return(COPY);

"+"				return(PLUS);
"-"				return(MINUS);
"*"				return(MULT);
"/"				return(DIV);
"%"				return(MOD);
"^"				return(XOR);
"|"				return(OR);
"&"				return(AND);
"~"				return(COMP);
"."				return(DOT);
"?"				return(QST);
":"				return(COL);
";"				return(';');
","				return(COMMA);
"{"				return('{');
"}"				return('}');
"["				return(LBRK);
"]"				return(RBRK);
"("				return('(');
")"				return(')');
"="				return(EQ);
"<"				return(LT);
">"				return(GT);
"!"				return(NOT);

^"#"{any}*	{
     		     String fp, fp2;
		     Import int abs();

		     /*
		      * A comment, line control statement or weird cpp op to be
		      * ignored. If it's a line control statement, change the
		      * internal line number and file name, otherwise punt it.
		      */
		     if (isdigit((*(yytext + 2))) && (fp=index(yytext, '"'))) {
			  yy_line = atonum(_currW, yytext + 1, 10);
			  if (yy_file) {
			       XtFree(yy_file); /* free old */
			       yy_file = NULL;
			  }
			  if (*(++fp) == '"') /* stdin */
			       yy_file = XtNewString("<stdin>");
			  else {
			       fp2 = index(fp, '"');
			       if (!fp2)
				    yyerror("Bad line control: %s", yytext);
			       else {
				    yy_file = XtMalloc((abs(fp2 - fp)) + 1);
				    strncpy(yy_file, fp, abs(fp2 - fp));
				    yy_file[abs(fp2 - fp)] = '\0';
			       }
			  }
		     }
		}

{sq-string}	{
		     if (*(yytext + 1) == '\\') {
			  char tmp[6];

			  strncpy(tmp, yytext + 1, 5);
			  tmp[5] = '\0';
			  backslash_eliminate(tmp, NORMAL_ELIMINATE, 0);
                          yylval.bval = *tmp;
			  return CHAR;
		     }
		     else if (strlen(yytext + 1) > 2) {
			  yyerror("Bad char constant %s", yytext);
			  return(0);
		     }
		     else {
			  yylval.bval = *(yytext + 1);
			  return CHAR;
		     }
		}

{string}	{
     		      /* nuke the quotes */
		      yylval.sval = XtMalloc(strlen(yytext) + 1);
		      strcpy(yylval.sval, yytext+1);
		      yylval.sval[strlen(yylval.sval) - 1] = '\0';
		      backslash_eliminate(yylval.sval, NORMAL_ELIMINATE, 0);
		      return STRING;
		}

{uniqueid}	{
		      static int unique_id = 0;
		      char foo[128];

		      sprintf(foo, "_uniq_%x_%x_id", _currW, unique_id++);
		      yylval.sval = XtNewString(foo);
		      return IDENTIFIER;
		}

{id}		{
		      int token;
		      /* is it a keyword like if, else, for, ... ? */
		      if (token = map_resword_to_token(yytext, T_KEYWORD))
			   return token;
		      /* how about a predefined type or class? */
		      else if (yylval.ival = map_resword_to_token(yytext,
								  T_TYPE |
								  T_CLASS))
			   return INT;
          	      else if ((yylval.cval = map_wname_to_wclass(yytext)) ||
			       (yylval.cval = lookup_client_wclass(_currW,
								   yytext)))
		           return WIDGET_TYPE;
		      else {
		           yylval.sval = XtNewString(yytext);
		           return IDENTIFIER;
		      }
		}

{hexnum}	{ yylval.ival = atonum(_currW, yytext, 16); return INT;	}

{octnum}	{ yylval.ival = atonum(_currW, yytext, 8);  return INT; }

{number}	{ yylval.ival = atonum(_currW, yytext, 10); return INT; }

{float}		{ yylval.fval = atof(yytext); return FLOAT;	}

{any}		{ yyerror("Invalid character '%c'", *yytext);	}
%%
