#ifndef lint
static char *rcsid = "$Header: /usr/src/local/awl/RCS/convert.c,v 2.4 90/04/19 20:04:48 jkh Exp $";
#endif

/*
 *
 *                     Copyright 1990
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * Various utilities for converting between types.
 *
 * $Log:	convert.c,v $
 * Revision 2.4  90/04/19  20:04:48  jkh
 * Alpha checkin.
 * 
 * Revision 2.3  90/03/31  06:54:44  jkh
 * Fixed bad error message in CALLBACK convert.
 * 
 * Revision 2.2  90/03/29  18:27:37  jkh
 * Fixed a lot more bugs with (PIXMAP) conversion.
 * 
 * Revision 2.1  90/03/27  08:58:56  jkh
 * Added PIXMAP conversion, moved do_convert() over from evalutils.c
 * 
 * Revision 2.0  90/03/26  01:44:15  jkh
 * pre-beta check-in
 * 
 */

#include "AwlP.h"
#include "y.tab.h"

/* This will lose on non-ascii architectures. I don't really care. */
#define DIGIT(x)        (isdigit(x) ? (x) - '0' : \
			 islower(x) ? (x) + 10 - 'a' : (x) + 10 - 'A')

Local char oct[] = "01234567";

/* determine whether a digit is valid for a given type/default radix */
Export Boolean isValidDigit(aw, ch, type)
AwlWidget aw;
char ch;
int type;
{
#ifndef toupper
     Import char toupper();
#endif

     /* sign char is valid for everyone */
     if (ch == '+' || ch == '-')
	  return(TRUE);
     /* next, check less restrictive float/any */
     else if ((ch == '.' || toupper(ch) == 'E' || isdigit(ch))
	 && (type == FLOAT || type == ANY))
	  return(TRUE);
     /* now check integer */
     else if (awl_radix(aw) == 8 ? (int)index(oct, ch) :
	      awl_radix(aw) == 16 ? isxdigit(ch) : isdigit(ch))
	  return(TRUE);
     return(FALSE);
}

/* Given a string, return an integer (doing any necessary base conversion) */
Export int atonum(aw, s, base)
AwlWidget aw;
char *s;
int base;
{
     int value = 0;
     int sign = -1;

     if (awl_radix(aw) == 0)
	  awl_radix(aw) = 10;

     if (base == 0)	/* set a default */
	  base = awl_radix(aw);

     if (!s || !strlen(s))
	  return(0);

     s = skip_whitespace(s);

     if (*s == '-') {
	  s++;
	  sign = 1;
     }
     else if (*s == '+') {
	  s++;
	  sign = -1;
     }
     /* skip the 0x digits */
     if (base == 16) {
	  if (!strncmp(s, "0x", 2) || !strncmp(s, "0X", 2))
	       s += 2;
	  else
	       return(0);
     }
     while (*s && ((base == 8 ? (int)index(oct, *s) : isdigit(*s))
		   || (base == 16 && isxdigit(*s)))) {
	  value = base * value - DIGIT(*s);
	  s++;
     }
     return(value * sign);
}

/*
 * Given a number, convert it to its string representation. Returned
 * string is static and should be saved by the caller.
 */
Export String numtoa(aw, i, base)
AwlWidget aw;
int i, base;
{
     Local char a[32];

     if (awl_radix(aw) == 0)
	  awl_radix(aw) = 10;

     if (base == 0)
	  base = awl_radix(aw);

     switch(base) {
     case 8:
	  sprintf(a, "%#o", i);
	  break;

     case 10:
	  sprintf(a, "%d", i);
	  break;

     case 16:
     default:
	  sprintf(a, "%#0x", i);
	  break;
     }
     return(a);
}

/*
 * Given a float, convert it to its string representation.
 */
Export String ftoa(f)
float f;
{
     Local char a[64];

     sprintf(a, "%g", f);
     return(a);
}

/*
 * This is the "top level" conversion routine. It handles type conversion
 * in one of three ways: If the type is one of awl's internal types (like
 * INT or FLOAT), it uses awl's internal coercion routine that's used for
 * implicit coercion operations. If type is not one of awl's internal
 * types but is one of the "special toolkit types", it handles it directly.
 * Finally, if it's nothing awl knows about, it calls another routine that
 * tries to convert it with the toolkit's type conversion mechanism.
 */
Export Value do_convert(aw, a1, a2)
AwlWidget aw;
Value a1, a2;
{
     Value res;
     int type, fail;

     /*
      * Conversion type can be specified in several ways:
      * If it's an INT type, then we presume that it's one of our
      * predefined internal types and try to deal with it accordingly.
      * If it's a string, it still may be one of our internal types
      * specified by its printname, so we try to look it up. If the string
      * wasn't a printname for one of our times (type was returned as zero)
      * we can then assume that it's the name of some Toolkit resource.
      */
     value_clear(res);

     if (value_type(a2) == INT) {
	  if (!(type = value_int(a2))) {
	       exec_warn(aw, "Attempt to convert to type 0, ignored.");
	       return(res);
	  }
     }
     else if (value_type(a2) == STRING)
	  type = map_resword_to_token(value_string(a2), T_TYPE);

     /* is conversion even necessary? */
     if (value_type(a1) == type)
	  res = a1;

     else switch (type) {
     case 0:
	  /*
	   * Presume it's a conversion to an unknown Toolkit type. Let
	   * convert_external() handle it.
	   */
	  value_set(res, DATA, ANY, any,
		    convert_external(aw, coerce(aw, a1, ANY),
				     value_string(a2), &fail));
	  if (fail)
	       exec_warn(aw, "Conversion failure for '%s'", value_string(a2));
	  break;

     case CALLBACK:
	  if (value_type(a1) != SYMBOL ||
	      symbol_class(value_symbol(a1)) != FUNCTION) {
	       exec_warn(aw, "Attempt to convert non-function to CALLBACK");
	       value_clear(res);
	  }
	  else {
	       Symbol tmp = value_symbol(a1);
	       XtCallbackRec *cb;

	       cb = (XtCallbackRec *)XtMalloc(sizeof(XtCallbackRec) * 2);
	       cb[1].callback = NULL;
	       cb[1].closure = 0;
	       switch(symbol_type(tmp)) {
	       case FUNC_BUILTIN:
	       case FUNC_USER:
		    cb[0].callback = (XtCallbackProc)do_callback_user;
		    cb[0].closure = (caddr_t)tmp;
		    value_set(res, DATA, ANY, any, (Generic)cb);
		    break;

	       case FUNC_CALLBACK:
		    cb[0].callback =
			 (XtCallbackProc)function_faddr(symbol_function(tmp));
		    cb[0].closure = function_data(symbol_function(tmp));
		    value_set(res, DATA, ANY, any, (Generic)cb);
		    break;

	       case FUNC_UNDEF:
		    exec_warn(aw, "Can't convert %s to CALLBACK; not defined",
			      symbol_name(tmp));
		    break;

	       default:
		    exec_warn(aw, "Can't convert function type %s to CALLBACK",
			      type_name(symbol_type(tmp)));
		    value_clear(res);
		    break;
	       }
	  }
	  break;

     case PIXMAP:
	  a1 = coerce(aw, a1, LIST);
	  if (!awl_ctable(aw))
	       exec_warn(aw, "No color table installed! No conversion done.");
	  else if (!list_is_rect(value_list(a1)))
	       exec_warn(aw, "Non-rectangular list for PIXMAP conversion.");
	  else if (strlen(value_list(a1)[0]) % ctab_nchars(awl_ctable(aw)))
	       exec_warn(aw, "List uneven for color table of %d chars/pixel.");
	  else {
	       Pixmap ret;
	       char cmap[32]; /* the largest we can expect, see xtkrtns.c */
	       int height, width, scanlen, nch, x, y;
	       String *l = value_list(a1);
	       CTable ct = awl_ctable(aw);
	       Pixel pix;
	       char *data;
	       XImage *img;

	       scanlen = strlen(l[0]);
	       nch = ctab_nchars(ct);
	       height = list_len(l);
	       width = scanlen / nch;
	       data = XtMalloc(height * width * 4);
	       ret = XCreatePixmap(XtDisplay(aw), XtWindow(aw) ? XtWindow(aw) :
				   RootWindowOfScreen(XtScreen(aw)),
				   width, height,
				   DefaultDepthOfScreen(XtScreen(aw)));
	       img = XCreateImage(XtDisplay(aw),
				  DefaultVisualOfScreen(XtScreen(aw)),
				  DefaultDepthOfScreen(XtScreen(aw)),
				  ZPixmap, 0, data, width, height, 32, 0);
	       for (y = 0; y < height; y++) {
		    for (x = 0; x < scanlen; x += nch) {
			 strncpy(cmap, l[y] + x, nch);
			 cmap[nch] = '\0';
			 pix = ctab_colors(ct)[(int)hash_search(ctab_hash(ct),
					       cmap, NULL, NULL)];
			 XPutPixel(img, x ? x / nch : 0, y, pix);
		    }
	       }
	       XPutImage(XtDisplay(aw), ret, DefaultGCOfScreen(XtScreen(aw)),
			 img, 0, 0, 0, 0, width, height);
	       XDestroyImage(img);
	       value_set(res, DATA, ANY, any, (Generic)ret);
	  }
	  break;

     case TRANSLATION:
	  a1 = coerce(aw, a1, ANY);
	  if (value_type(a1) != STRING) {
	       exec_warn(aw, "Attempt to convert non-string to TRANSLATION");
	       value_clear(res);
	  }
	  else {
	       Local Boolean added_action = FALSE;
	       int cnt, func;
	       String src, dst;
	       char fname[256];
	       XtTranslations trans;

	       if (!added_action) {
		    Import void do_action_user();
		    Local XtActionsRec do_action[] =
			 { { "DOAWL", do_action_user } };

		    XtAppAddActions(XtWidgetToApplicationContext(aw),
				    do_action, XtNumber(do_action));
		    added_action = TRUE;
	       }
	       /* first, try to figure out how much space we'll need */
	       cnt = strlen(value_string(a1));
	       src = value_string(a1);
	       while (*src != '\0')
		    if (*(src++) == ':')
			 cnt += 6;
	       /*
		* Now copy the translations, converting them on-the-fly.
		* This entails looking for all the "function" references
		* in the translation table and converting them to function
		* calls to the DOAWL() action. The original function name
		* becomes the first parameter to DOAWL() with the original
		* parameters (if any) being forwarded by DOAWL() to the
		* real target function. If the function name is not known
		* to awl, however, or is known to be already an ACTION
		* function, we leave it alone.
		*/
	       debug(aw, "Translations before convert: %s", value_string(a1));
	       src = value_string(a1);
	       dst = XtMalloc(cnt + 1);
	       cnt = 0;
	       while (*src != '\0') {
		    dst[cnt++] = *src;
		    if (*(src++) == ':') {
			 src = skip_whitespace(src);
			 func = 0;
			 while (*src && *src != '(')
			      fname[func++] = *(src++);
			 if (*(src++) != '(')
			      exec_warn(aw, "Bad TRANSLATION, missing '('");
			 else {
			      Symbol tmp;

			      fname[func] = '\0';
			      tmp = symbol_find(aw, fname);
			      if (!tmp || symbol_type(tmp) == FUNC_ACTION
				  || symbol_class(tmp) != FUNCTION) {
				   strcpy(dst + cnt, fname);
				   cnt += func;
				   dst[cnt++] = '(';
			      }
			      else {
				   strcpy(dst + cnt, "DOAWL(");
				   cnt += 6;
				   strcpy(dst + cnt, fname);
				   cnt += func;
				   src = skip_whitespace(src);
				   if (*src != ')')
					dst[cnt++] = ',';
			      }
			 }
		    }
	       }
	       dst[cnt] = '\0';
 	       debug(aw, "Translations after convert: %s", dst);
	       trans = XtParseTranslationTable(dst);
	       XtFree(dst);
	       value_set(res, DATA, ANY, any, (Generic)trans);
          }
	  break;

     case WIDGET:
	  a1 = coerce(aw, a1, ANY);
	  if (value_type(a1) == LAYOUT) {
	       if (wl_id(value_any(a1)))
		    value_set(res, DATA, WIDGET, widget, wl_id(value_any(a1)));
	       else
		    value_set(res, INDIRECT, WIDGET, any,
			      (Generic)&wl_id(value_any(a1)));
	  }
	  else /* assume they know what they're doing */
	       value_set(res, DATA, WIDGET, any, value_any(a1));
	  break;

     case LIST:	/* special pre-prep case for LIST */
	  if (value_type(a1) == SYMBOL
	      && symbol_type(value_symbol(a1)) != STRING
	      && symbol_alist(value_symbol(a1))) {
	       String *tmp;

	       tmp = list_from_alist(aw, symbol_alist(value_symbol(a1)));
	       value_set(res, DATA, LIST, aobj, new_aobj(tmp));
	       break;
	  }

     default:
	  /* handle it as a standard known type */
	  res = coerce(aw, a1, type);
	  break;
     }
     return(res);
}

/*
 *
 * DANGER Will Robinson! XtConvert() hands back an address which
 * may be used in different ways depending on what we've converted.
 * We try to deal with this as best possible, but perhaps not successfully
 * in all cases. We use a fail flag to indicate failure since the returned
 * value can be zero.
 */
Export Generic convert_external(aw, val, to_type, fail)
AwlWidget aw;
Value val;
String to_type;
int *fail;
{
     XrmValue from, to;

     to.addr = NULL;
     to.size = 0;
     *fail = 0;	/* Be optimistic */

     val = coerce(aw, val, ANY);

     switch (value_type(val)) {
     case STRING:
	  from.addr = value_string(val);
	  from.size = strlen(value_string(val)) + 1;
	  XtConvert(aw, "String", &from, to_type, &to);
	  break;
	  
     case INT:
	  from.addr = (Generic)&(value_int(val));
	  from.size = sizeof(int);
	  XtConvert(aw, "Int", &from, to_type, &to);
	  break;
	  
     case FLOAT:
	  from.addr = (Generic)&(value_float(val));
	  from.size = sizeof(FTYPE);
	  XtConvert(aw, "Float", &from, to_type, &to);
	  break;
	  
     default:
	  to.size = 0;
	  break;
     }
     if (to.size == 0 || to.addr == NULL) {
	  *fail = 1;
	  return(NULL);
     }
     if (to.size == 1)
	  return((Generic)(*to.addr));
     else if (to.size == 2)
	  return((Generic)(*((XtArgVal *)to.addr)));
     else
	  return((Generic)(*(long *)to.addr));
}

/* Coerce val to type, returning new value */
Export Value coerce(aw, val, type)
AwlWidget aw;
Value val;
int type;
{
     Value ret;

     if (value_type(val) == type)
	  return(val);
     else
	  val = do_load(aw, val, FALSE);

     if (value_type(val) == type || type == ANY)
	  return(val);

     value_type(ret) = type;
     value_class(ret) = value_class(val);	/* retain class */

     switch (type) {
     case CHAR:
	  switch(value_type(val)) {
	  case INT:
	       value_char(ret) = value_int(val);
	       break;

	  case FLOAT:
	       value_char(ret) = (char)value_float(val);
	       break;

	  case STRING:
	       value_char(ret) = (char)atonum(aw, value_string(val), 0);
	       break;

	  default:
	       value_char(ret) = (char)value_any(val);
	       break;
	  }
	  break;

     case INT:
	  switch(value_type(val)) {
	  case CHAR:
	       value_int(ret) = value_char(val);
	       break;

	  case FLOAT:
	       value_int(ret) = (int)value_float(val);
	       break;

	  case STRING:
	       value_int(ret) = atonum(aw, value_string(val), 0);
	       break;

	  default:
	       value_int(ret) = (int)value_any(val);
	       break;
	  }
	  break;

     case FLOAT:
	  switch(value_type(val)) {
	  case CHAR:
	       value_float(ret) = (FTYPE)value_char(val);
	       break;

	  case INT:
	       value_float(ret) = (FTYPE)value_int(val);
	       break;

	  case STRING:
	       value_float(ret) = (FTYPE)atonum(aw, value_string(val), 0);
	       break;

	  default:
	       value_any(ret) = value_any(val);
	       break;
	  }
	  break;

     case STRING:
	  switch(value_type(val)) {
	       String str;

	  case CHAR:
	       str = XtMalloc(2);
	       str[0] = value_char(val);
	       str[1] = '\0';
	       value_aobj(ret) = new_aobj(str);
	       break;

	  case INT:
	       str = numtoa(aw, value_int(val), 0);
	       value_aobj(ret) = new_aobj(XtNewString(str));
	       break;

	  case FLOAT:
	       str = ftoa(value_float(val));
	       value_aobj(ret) = new_aobj(XtNewString(str));
	       break;

	  case LIST:
	       value_aobj(ret) = new_aobj(list_to_string(aw, value_list(val)));
	       break;

	  case WCLASS:
	       if ((str = map_wclass_to_wname(value_wclass(val))) != NULL)
		    value_aobj(ret) = new_aobj(XtNewString(str));
	       else
		    value_clear(ret);
	       break;

	  default:
	       value_class(ret) = STATIC;
	       value_aobj(ret) = new_aobj(value_any(val));
	       break;
	  }
	  break;

     case LIST:
	  switch(value_type(val)) {

	  case CHAR:
	       value_aobj(ret) = new_aobj(list_from_string(aw, numtoa(aw,
							     value_char(val),
							     0)));
	       break;

	  case INT:
	       value_aobj(ret) = new_aobj(list_from_string(aw, numtoa(aw,
							     value_int(val),
							     0)));
	       break;

	  case FLOAT:
	       value_aobj(ret) =
		    new_aobj(list_from_string(aw, ftoa(value_float(val))));
	       break;

	  case STRING:
	       value_aobj(ret) = 
		    new_aobj(list_from_string(aw, value_string(val)));
	       break;

	  default:
	       value_class(ret) = STATIC;
	       value_aobj(ret) = new_aobj(value_any(val));
	       break;
	  }
	  break;

     case WCLASS:
	  if (value_type(val) != STRING) {
	       exec_warn("cast: Attempt to coerce type '%s' to WidgetClass",
			 type_name(value_type(val)));
	       value_clear(ret);
	  }
	  else
	       value_wclass(ret) = map_wname_to_wclass(value_string(val));
	  break;

     default:
	  value_any(ret) = value_any(val);
	  break;
     }
     return(ret);
}
