#ifndef _MACROS_H_INCLUDE
#define _MACROS_H_INCLUDE

/* $Header: /usr/src/local/awl/RCS/macros.h,v 2.4 90/04/19 20:05:22 jkh Exp $ */

/*
 *
 *                   Copyright 1989, 1990
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * Here are all the various and sundry macros that make life easier.
 *
 * $Log:	macros.h,v $
 * Revision 2.4  90/04/19  20:05:22  jkh
 * Alpha checkin.
 * 
 * Revision 2.3  90/03/31  22:17:55  jkh
 * None.
 * 
 * Revision 2.2  90/03/27  08:33:01  jkh
 * Some general cleanup.
 * 
 * Revision 2.1  90/03/27  08:00:26  jkh
 * Added evil XtColormap macro. See comment.
 * 
 * Revision 2.0  90/03/26  01:43:26  jkh
 * pre-beta check-in
 * 
 */

/* General utility macros */

#define Local static
#define Import extern
#define Export

#if defined(_GNUC_) && !defined(NOINLINE)
#define Inline inline
#else
#define Inline
#endif

#ifndef TRUE	/* someday these will be someplace reliable (posix?) */
#define TRUE	(1)
#define FALSE	(0)
#endif

/* What version are we? */
#define VERSION		0

#define UNDEF_ADDR	-1

/* For signals */
#define MY_SIG_DFL	0
#define MY_SIG_IGN	1

/* define the size of global and "alist" hash tables */
#define GLOBAL_HASH_SIZE	97
#define ALIST_HASH_SIZE		37

/* For backslash_eliminate. This has to stay in sync with strsed.c. Puke. */
#define NORMAL_ELIMINATE	2

/* number of saved registers in stack frame */
#define FRAME_REG_SIZE	3

/* how to get to a given argument via the frame pointer */
#define FRAME_OFFSET(parm) \
     ((parm) < 0 ? (awl_fp(aw) - (awl_nparms(aw) + (parm) + FRAME_REG_SIZE)) \
      : (awl_fp(aw) + (parm)))

/* FILEPXDESC macros */
#define PX_IN		0
#define PX_OUT		1

#ifndef NFILES		/* this is usually hard to find */
#define NFILES		64
#endif

#ifndef MAXSIG
#define MAXSIG		32
#endif

/*
 * All these are power-of-2 - 8 so that malloc() doesn't allocate
 * two blocks instead of one (as it would if the amount was already
 * at the edge of a boundry). In practice, this doesn't usually happen
 * with values < 1K, but some malloc routines allocate block lists for
 * the smaller requests too.
 */
#define TEXT_INCR	1016	/* text segment increment size */
#define LTAB_INCR	248	/* line table increment size */
#define STACK_INCR	56	/* initial stack segment size */
#define ALLOC_INCR	56	/* allocation tick */

/* For mapping widget/resource names in gperf files */
#define DEFRMAP(name) \
     Local void name(aw, args, idx, res) \
     AwlWidget aw; Arg *args; int *idx; Resource res;

#define DEFWMAP(name) \
     Local void name(n, class, parent, args, nargs) \
     String n; WidgetClass class; Widget parent; Arg *args; int nargs;

#define MAP(name) ((WidgetClass *)(name))


#if defined(__STDC__) && !defined(UNIXCPP)	/* ANSI */
#define DEFUN(name)	Export Value awl_##name(aw) \
			AwlWidget aw;

#define F_DECLARE(awl_widget, name, args) \
     { \
     extern Value awl_##name(); \
     addr_fp(addr) = awl_##name; \
     symbol_def_function(awl_widget, #name, FUNC_BUILTIN, args, addr, FALSE); \
     }

#define C_DECLARE(awl_widget, name, type, val) \
     symbol_def_constant(awl_widget, #name, type, val)

#define V_DECLARE(awl_widget, name, type, val) \
     symbol_def_variable(awl_widget, #name, type, val)

#else
#define DEFUN(name)	Export Value awl_/**/name(aw) \
     			AwlWidget aw;

#define F_DECLARE(awl_widget, name, args) \
     { \
     extern Value awl_/**/name(); \
     addr_fp(addr) = awl_/**/name; \
     symbol_def_function(awl_widget, "name", FUNC_BUILTIN, args, addr, FALSE);\
     }

#define C_DECLARE(awl_widget, name, type, val) \
     symbol_def_constant(awl_widget, "name", type, val)

#define V_DECLARE(awl_widget, name, type, val) \
     symbol_def_variable(awl_widget, "name", type, val)

#endif

/* Define this as the type of float you want to use */
#define FTYPE		float

#ifndef MAXFLOAT
#define MAXFLOAT ((float)3.40282346638528860e+38)
#endif

/* For the POP operator */
#define DISCARD		0
#define NODUP		1
#define DUPVAL		2
#define DUPTARG		3

/*
 * Now we get to all sorts of "accessor" macros. Yeah, I'm on a Common Lisp
 * kick, but that's beside the point. None of AWL's internal data types are
 * ever accessed directly; everything going through these macros in order
 * to make radical changes in awl's internal structure a lot easier.
 * This approach has already saved my bacon about 3 times!
 *
 * (P.S. If you think these macros are hairy, hah. Look at a Common Lisp
 * loop macro sometime).
 */

/* Process accessors */
#define process_size		(sizeof(struct _process))
#define proc_pid(proc)		(((Process)(proc))->pid)
#define proc_in(proc)		(((Process)(proc))->in)
#define proc_out(proc)		(((Process)(proc))->out)

/* Allocated object accessors */
#define aobj_size		(sizeof(struct _allocobj))
#define aobj_cnt(ao)		(((AllocObj)(ao))->cnt)
#define aobj_ptr(ao)		(((AllocObj)(ao))->ptr)
#define aobj_string(ao)		((String)aobj_ptr(ao))
#define aobj_list(ao)		((String *)aobj_ptr(ao))

/* Address accessors */
#define addr_size		(sizeof(struct _addr))
#define addr_fp(adr)		((adr).s)
#define addr_ep(adr)		((adr).f)
#define addr_pc(adr)		((adr).u)

/* Datum accessors */
#define datum_size		(sizeof(struct _datum))
#define datum_any(dat)		((dat).a)
#define datum_aobj(dat)		((dat).o)
#define datum_char(dat)		((dat).b)
#define datum_file(dat)		((dat).z)
#define datum_float(dat)	((dat).f)
#define datum_int(dat)		((dat).i)
#define datum_range(dat)	((dat).r)
#define datum_symbol(dat)	((dat).y)
#define datum_wclass(dat)	((dat).c)
#define datum_widget(dat)	((dat).w)
#define datum_list(dat)		(aobj_list(datum_aobj(dat)))
#define datum_string(dat)	(aobj_string(datum_aobj(dat)))

/* Range value accessors */
#define range_min(rng)		((rng)[0])
#define range_max(rng)		((rng)[1])

/* Value accessors */
#define value_size		(sizeof(struct _val))
#define value_addr(val)		(&(val))
#define value_class(val)	((val).class)
#define value_data(val)		((val).v)
#define value_type(val)		((val).type)
#define value_any(val)		(datum_any(value_data(val)))
#define value_aobj(val)		(datum_aobj(value_data(val)))
#define value_char(val)		(datum_char(value_data(val)))
#define value_file(val)		(datum_file(value_data(val)))
#define value_float(val)	(datum_float(value_data(val)))
#define value_function(val)	((Function)value_any(val))
#define value_int(val)		(datum_int(value_data(val)))
#define value_list(val)		(datum_list(value_data(val)))
#define value_range(val)	(datum_range(value_data(val)))
#define value_string(val)	(datum_string(value_data(val)))
#define value_symbol(val)	(datum_symbol(value_data(val)))
#define value_wclass(val)	(datum_wclass(value_data(val)))
#define value_widget(val)	(datum_widget(value_data(val)))

#if defined(__STDC__) && !defined(UNIXCPP)	/* ANSI */
#define value_set(val, class, type, member, vv) \
     ((void)(value_class(val) = (class), value_type(val) = (type), \
     value_##member(val) = (vv)))
#else
#define value_set(val, class, type, member, vv) \
     ((void)(value_class(val) = (class), value_type(val) = (type), \
     value_/**/member(val) = (vv)))
#endif

#define value_clear(arg)	value_set((arg), DATA, INT, int, 0)

/* Symbol accessors */
#define symbol_size		(sizeof(struct _sym))
#define symbol_root(sym)	(((Symbol)(sym))->aw)
#define symbol_name(sym)	(((Symbol)(sym))->name)
#define symbol_alist(sym)	(((Symbol)(sym))->alist)
#define symbol_parent(sym)	(((Symbol)(sym))->parent)
#define symbol_value(sym)	(((Symbol)(sym))->v)
#define symbol_next(sym)	(((Symbol)(sym))->next)

#define symbol_addr(sym)	(value_addr(symbol_value(sym)))
#define symbol_any(sym)		(value_any(symbol_value(sym)))
#define symbol_aobj(sym)	(value_aobj(symbol_value(sym)))
#define symbol_char(sym)	(value_char(symbol_value(sym)))
#define symbol_class(sym)	(value_class(symbol_value(sym)))
#define symbol_data(sym)	(value_data(symbol_value(sym)))
#define symbol_file(sym)	(value_file(symbol_value(sym)))
#define symbol_float(sym)	(value_float(symbol_value(sym)))
#define symbol_function(sym)	(value_function(symbol_value(sym)))
#define symbol_int(sym)		(value_int(symbol_value(sym)))
#define symbol_list(sym)	(value_list(symbol_value(sym)))
#define symbol_range(sym)	(value_range(symbol_value(sym)))
#define symbol_string(sym)	(value_string(symbol_value(sym)))
#define symbol_symbol(sym)	(value_symbol(symbol_value(sym)))
#define symbol_type(sym)	(value_type(symbol_value(sym)))
#define symbol_wclass(sym)	(value_wclass(symbol_value(sym)))
#define symbol_widget(sym)	(value_widget(symbol_value(sym)))

#define is_allocated_type(typ)	((typ) == STRING || (typ) == LIST)

#define is_filedesc_type(typ)	((typ) == FILEDESC || (typ) == FILEPDESC)

#define is_function_type(typ)	((typ)==FUNC_EXTERN || (typ)==FUNC_ACTION || \
				 (typ)==FUNC_CALLBACK || (typ)==FUNC_USER || \
				 (typ)==FUNC_BUILTIN)

#define is_precious(class)	((class) == CONSTANT || (class) == ALLOCATED)

#define is_register(typ)	((typ) == _PC || (typ) == _SP || (typ) == _FP)

#define AOBJCASE		case STRING: case LIST

#define symbol_is_char(sym)	(symbol_type(sym) == CHAR)
#define symbol_is_float(sym)	(symbol_type(sym) == FLOAT)
#define symbol_is_int(sym)	(symbol_type(sym) == INT)
#define symbol_is_list(sym)	(symbol_type(sym) == LIST)
#define symbol_is_string(sym)	(symbol_type(sym) == STRING)
#define symbol_is_allocated(sym)(symbol_class(sym) == ALLOCATED)
#define symbol_is_constant(sym)	(symbol_class(sym) == CONSTANT)
#define symbol_is_function(sym)	(symbol_class(sym) == FUNCTION)
#define symbol_is_label(sym)	(symbol_class(sym) == LABEL)
#define symbol_is_static(sym)	(symbol_class(sym) == STATIC)
#define symbol_is_precious(sym)	(is_precious(symbol_class(sym)))
#define symbol_is_register(sym)	(is_register(symbol_type(sym)))

/* Function accessors */
#define function_size		(sizeof(struct _function))
#define function_addr(func)	(((Function)(func))->addr)
#define function_uaddr(func)	(addr_pc(function_addr(func)))
#define function_faddr(func)	(addr_fp(function_addr(func)))
#define function_eaddr(func)	(addr_ep(function_addr(func)))
#define function_nparms(func)	(((Function)(func))->nparms)
#define function_nlocals(func)	(((Function)(func))->nlocals)
#define function_data(func)	(((Function)(func))->calldata)
#define function_end(func)	(((Function)(func))->end)

/* Widget layout accessors */
#define wl_size			(sizeof(struct _layout))
#define wl_id(wl)		(((Layout)(wl))->w_id)
#define wl_name(wl)		(((Layout)(wl))->name)
#define wl_wclass(wl)		(((Layout)(wl))->class)
#define wl_parent(wl)		(((Layout)(wl))->parent)
#define wl_head(wl)		(((Layout)(wl))->head)
#define wl_tail(wl)		(((Layout)(wl))->tail)
#define wl_prev(wl)		(((Layout)(wl))->prev)
#define wl_next(wl)		(((Layout)(wl))->next)
#define wl_resources(wl)	(((Layout)(wl))->res)

/* Resource accessors */
#define res_size		(sizeof(struct _resource))
#define res_name(res)		(((Resource)(res))->name)
#define res_value(res)		(((Resource)(res))->val)
#define res_next(res)		(((Resource)(res))->next)
#define res_parent(res)		(((Resource)(res))->parent)

/* Line table accessors */
#define ltab_size		(sizeof(struct _ltabent))
#define ltab_file(ltab)		((ltab).file)
#define ltab_line(ltab)		((ltab).line)
#define ltab_start(ltab)	((ltab).start)
#define ltab_last(ltab)		((ltab).last)

/* Breakpoint accessors */
#define bkpt_size		(sizeof(struct _breakpoint))
#define bkpt_next(bk)		(((Breakpoint)(bk))->next)
#define bkpt_func(bk)		(((Breakpoint)(bk))->func)
#define bkpt_addr(bk)		(((Breakpoint)(bk))->addr)
#define bkpt_enable(bk)		(((Breakpoint)(bk))->enable)

/* Color table accessors */
#define ctab_size		(sizeof(struct _ctable))
#define ctab_colors(ctab)	(((CTable)(ctab))->colors)
#define ctab_hash(ctab)		(((CTable)(ctab))->htab)
#define ctab_nchars(ctab)	(((CTable)(ctab))->nch)
#define ctab_ncolors(ctab)	(((CTable)(ctab))->ncolors)

/* Widget accessors */
#define awl_active(aw)		(((AwlWidget)(aw))->awl.compile.active)
#define awl_bkpts(aw)		(((AwlWidget)(aw))->awl.debuginfo.bkpts)
#define awl_bomb(aw)		(((AwlWidget)(aw))->awl.sysinfo.bomb_out)
#define awl_cpp_args(aw)	(((AwlWidget)(aw))->awl.res.cpp_args)
#define awl_cpp_cmd(aw)		(((AwlWidget)(aw))->awl.res.cpp_command)
#define awl_ctable(aw)		(((AwlWidget)(aw))->awl.sysinfo.ctable)
#define awl_current(aw)		(((AwlWidget)(aw))->awl.compile.current)
#define awl_debug(aw)		(((AwlWidget)(aw))->awl.res.debug)
#define awl_err(aw)		(((AwlWidget)(aw))->awl.io.err)
#define awl_estack(aw)		(((AwlWidget)(aw))->awl.seginfo.estack)
#define awl_etext(aw)		(((AwlWidget)(aw))->awl.seginfo.etext)
#define awl_externlist(aw)	(((AwlWidget)(aw))->awl.res.externlist)
#define awl_file(aw)		(((AwlWidget)(aw))->awl.res.file)
#define awl_fp(aw)		(((AwlWidget)(aw))->awl.segptrs.fp)
#define awl_in(aw)		(((AwlWidget)(aw))->awl.io.in)
#define awl_lastinsn(aw)	(((AwlWidget)(aw))->awl.compile.lastinsn)
#define awl_lastpc(aw)		(((AwlWidget)(aw))->awl.compile.lastpc)
#define awl_lineinfo(aw)	(((AwlWidget)(aw))->awl.res.lineinfo)
#define awl_ltab(aw)		(((AwlWidget)(aw))->awl.debuginfo.ltab)
#define awl_ltabcurr(aw)	(((AwlWidget)(aw))->awl.debuginfo.ltabcurr)
#define awl_ltabsize(aw)	(((AwlWidget)(aw))->awl.debuginfo.ltabsize)
#define awl_mainargv(aw)	(((AwlWidget)(aw))->awl.res.argv)
#define awl_nowarn(aw)		(((AwlWidget)(aw))->awl.res.nowarn)
#define awl_nparms(aw)		(((AwlWidget)(aw))->awl.segptrs.nparms)
#define awl_oldsp(aw)		(((AwlWidget)(aw))->awl.segptrs.oldsp)
#define awl_out(aw)		(((AwlWidget)(aw))->awl.io.out)
#define awl_pc(aw)		(((AwlWidget)(aw))->awl.segptrs.pc)
#define awl_radix(aw)		(((AwlWidget)(aw))->awl.io.radix)
#define awl_sep(aw)		(((AwlWidget)(aw))->awl.io.sep)
#define awl_showasm(aw)		(((AwlWidget)(aw))->awl.res.showasm)
#define awl_sp(aw)		(((AwlWidget)(aw))->awl.segptrs.sp)
#define awl_stack(aw)		(((AwlWidget)(aw))->awl.seginfo.stack)
#define awl_startfunc(aw)	(((AwlWidget)(aw))->awl.res.startfunc)
#define awl_text(aw)		(((AwlWidget)(aw))->awl.seginfo.text)
#define awl_top(aw)		(((AwlWidget)(aw))->awl.sysinfo.top)
#define awl_trace(aw)		(((AwlWidget)(aw))->awl.res.trace)
#define awl_version(aw)		(((AwlWidget)(aw))->awl.res.version)

/* These are the flag bits for widget maps */
#define W_NONE			0x0
#define W_EXTENDED		0x1
#define W_POPUP			0x2
#define W_MANAGER		0x4
#define W_RESPROC		0x8
#define W_WIDPROC		0x10

/* These are the flag bits for reserved word maps */
#define T_NONE			0x0
#define T_KEYWORD		0x1
#define T_INTERN_TYPE		0x2
#define T_TYPE			0x4
#define T_CLASS			0x8
#define T_ALL			0xFF

#endif /* _MACROS_H_INCLUDE */
