#ifndef lint
static char *rcsid = "$Header: /usr/src/local/awl/RCS/symbols.c,v 2.4 90/04/19 20:05:51 jkh Exp $";
#endif

/*
 *
 *                     Copyright 1989
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * This file contains all the functions for manipulating "symbols" in awl.
 * Each symbol can contain its own table of symbols (alists) with a special
 * purpose symbol alist (Symbol_top) being used to implement the global
 * symbol table.
 *
 * $Log:	symbols.c,v $
 * Revision 2.4  90/04/19  20:05:51  jkh
 * Alpha checkin.
 * 
 * Revision 2.3  90/03/31  06:20:35  jkh
 * Fixed Yet Another Typo.
 * 
 * Revision 2.2  90/03/31  05:51:56  jkh
 * Fixed typo.
 * 
 * Revision 2.1  90/03/31  05:49:45  jkh
 * Changed the way register symbols are printed.
 * 
 * Revision 2.0  90/03/26  01:43:50  jkh
 * pre-beta check-in
 *
 */

#include "AwlP.h"
#include "y.tab.h"

Local Symbol LocalParmList;

Local int symbol_free_contents();
Local void symbol_free();

/* Return a terse description of a symbol */
Export String symbol_printname(sym)
Symbol sym;
{
     Local char prtname[256];

     /* limit it to 200 just to be safe, however unlikely */
     strncpy(prtname, symbol_name(sym), 200);
     prtname[200] = '\0';
     if (symbol_is_register(sym)) {
	  strcat(prtname, ":");
	  strcat(prtname, type_name(symbol_type(sym)));
	  sprintf(prtname + strlen(prtname), "%+d", symbol_int(sym));
     }
     return(prtname);
}

/* Allocate space for symbol named "name" of type "type". */
Export Inline Symbol symbol_new(aw, name, type)
AwlWidget aw;
String name;
int type;
{
     Symbol ret;

     ret = (Symbol)XtMalloc(symbol_size);
     bzero(ret, symbol_size);
     symbol_name(ret) = XtNewString(name);
     /* some reasonable defaults */
     symbol_root(ret) = aw;
     symbol_class(ret) = DATA;
     symbol_type(ret) = type;
     symbol_parent(ret) = awl_top(aw);
     return(ret);
}

/*
 * Free just the storage for a symbol's contents, returning 1 if this
 * is possible, 0 if not.
 */
Local Inline int symbol_free_contents(sym)
Symbol sym;
{
     /*
      * If it's an allocated type, then the contents were malloc'd
      * and should be free'd, unless the symbol is precious.
      */
     if (is_allocated_type(symbol_type(sym)) && !symbol_is_precious(sym)) {
	  do_free(&symbol_value(sym));
	  /* If it has an alist, free it */
	  if (symbol_alist(sym))
	       hash_purge(symbol_alist(sym), symbol_free);
	  return(1);
     }
     else
	  return(0);
}
	  
/* Free a symbol's storage. */
Local Inline void symbol_free(sym)
Symbol sym;
{
     if (symbol_free_contents(sym)) {
	  debug(symbol_root(sym), "symbol_free: nuking %s\n",
		symbol_name(sym));
	  XtFree(symbol_name(sym));
	  XtFree(sym);
     }
}
     
/* Add a symbol to a symbol table. */
Export Inline void symbol_add(sym)
Symbol sym;
{
     debug(symbol_root(sym), "symbol_add: Adding %s to table %0#x",
	   symbol_name(sym), symbol_parent(sym));
     hash_search(symbol_parent(sym), symbol_name(sym), sym,
		 symbol_free_contents);
}

Export Inline void symbol_set_locals(s)
Symbol s;
{
     Symbol tmp = LocalParmList;

     if (s) {
	  if (tmp) {
	       while (symbol_next(tmp))
		    tmp = symbol_next(tmp);
	       symbol_next(tmp) = s;
	  }
	  else 
	       LocalParmList = s;
     }
     else
	  LocalParmList = NULL;
}

/* Remove a symbol from a symbol table and free it. */
Export Inline void symbol_remove(aw, sym)
AwlWidget aw;
Symbol sym;
{
     debug(aw, "symbol_remove: Removing %s from table %0#x",
	   symbol_name(sym), symbol_parent(sym));
     hash_destroy(symbol_parent(sym), symbol_name(sym), symbol_free);
}

/* Find a symbol in the symbol table. */
Export Symbol symbol_find(aw, name)
AwlWidget aw;
String name;
{
     Symbol ret;

     if (!name)
	  return(NULL);

     if (LocalParmList) {	/* search these first */
	  Symbol tmp = LocalParmList;
	  
	  while (tmp) {
	       if (!strcmp(symbol_name(tmp), name))
		    return(tmp);
	       else
		    tmp = symbol_next(tmp);
	  }
     }
     ret = (Symbol)hash_search(awl_top(aw), name, NULL, NULL);
     return(ret);
}

/* look to see if a given name is a special widget class */
Export Inline WidgetClass lookup_client_wclass(aw, name)
AwlWidget aw;
String name;
{
     Symbol tmp;

     tmp = symbol_find(aw, name);
     if (tmp && symbol_class(tmp) == CONSTANT && symbol_type(tmp) == WCLASS)
	  return(symbol_wclass(tmp));
     else
	  return((WidgetClass)NULL);
}

/* Find a value in another symbol's alist */
Export Symbol symbol_alist_find(sym, val, addp)
Symbol sym;
Value val;
int addp;
{
     Symbol ret;
     String name;
     AwlWidget aw;

     aw = symbol_root(sym);
     name = value_string_value(aw, val);
     if (!symbol_alist(sym)) {
	  if ((symbol_alist(sym) = hash_create(ALIST_HASH_SIZE)) == NULL)
	       exec_error(aw, "Couldn't create hash table for alist!");
     }
     /* look for it */

     ret = (Symbol)hash_search(symbol_alist(sym), name, NULL, NULL);
     debug(aw, "alist_find: Search for %s returns %d", name, ret);
     if (addp && !ret) {
	  ret = symbol_new(aw, name, STRING);
	  value_set(symbol_value(ret), DATA, STRING, aobj,
		    new_aobj(XtNewString(name)));
	  symbol_parent(ret) = symbol_alist(sym);
	  hash_search(symbol_alist(sym), name, ret, NULL);
     }
     return(ret);
}

/* Add a function to the symbol table, declaring it undefined. */
Export Inline Symbol symbol_add_function(aw, n)
AwlWidget aw;
String n;
{
     Symbol tmp;

     tmp = symbol_new(aw, n, FUNC_UNDEF);
     symbol_class(tmp) = FUNCTION;
     symbol_function(tmp) = (Function)XtMalloc(function_size);
     function_uaddr(symbol_function(tmp)) = UNDEF_ADDR;
     function_nparms(symbol_function(tmp)) =
	  function_nlocals(symbol_function(tmp)) = 0;
     symbol_add(tmp);
     return(tmp);
}

/* Define function in the symbol table. */
Export Symbol symbol_def_function(aw, n, type, nparms, fp, checkp)
AwlWidget aw;
String n;
int type, nparms;
Address fp;
Boolean checkp;
{
     Symbol tmp;
     Function f;

     /* See if it was already allocated through a forward ref */
     if (checkp && (tmp = symbol_find(aw, n)) != NULL) {
	  if (!symbol_is_function(tmp)) {
	       exec_error(aw, "Attempt to redefine symbol '%s' as function.",
			  symbol_name(tmp));
	       return(NULL);
	  }
	  if (symbol_type(tmp) != FUNC_UNDEF) {
	       exec_error(aw, "'%s' is already defined as a %s function.",
			  symbol_type(tmp) == FUNC_BUILTIN ? "system" : "user",
			  symbol_name(tmp));
	       return(NULL);
	  }
     }
     else
	  tmp = symbol_add_function(aw, n);
     symbol_type(tmp) = type;
     f = symbol_function(tmp);
     function_addr(f) = fp;
     function_nparms(f) = nparms;
     return(tmp);
}

/* define a constant in the symbol table */
Export void symbol_def_constant(aw, n, type, val)
AwlWidget aw;
String n;
int type;
Datum val;
{
     Symbol tmp;

     if (type) {
	  tmp = symbol_new(aw, n, type);
	  symbol_class(tmp) = CONSTANT;
	  symbol_data(tmp) = val;
     }
     else {
	  /* it's a string name constant */
	  tmp = symbol_new(aw, n, STRING);
	  symbol_class(tmp) = CONSTANT;
	  symbol_aobj(tmp) = new_aobj(n);
     }
     symbol_add(tmp);
}

/* define a user-setable variable in the symbol table */
Export void symbol_def_variable(aw, n, type, val)
AwlWidget aw;
String n;
int type;
Generic val;
{
     Symbol tmp;

     tmp = symbol_new(aw, n, type);
     symbol_class(tmp) = DATA;
     symbol_any(tmp) = val;
     symbol_add(tmp);
}
