#ifndef lint
static char *rcsid = "$Header: /usr/src/local/awl/RCS/sysrtns.c,v 2.1 90/04/19 20:05:56 jkh Exp $";
#endif

/*
 *
 *                     Copyright 1989
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * This file contains all the system (UNIX) specific built-ins and
 * a few awl utility functions.
 *
 * Note that mathrtns.c also uses some UNIX dependant features
 * to handle floating point traps.
 *
 * $Log:	sysrtns.c,v $
 * Revision 2.1  90/04/19  20:05:56  jkh
 * Alpha checkin.
 * 
 * Revision 2.0  90/03/26  01:43:52  jkh
 * pre-beta check-in
 * 
 */

#include "AwlP.h"
#include "y.tab.h"
#include <dirent.h>
#include <grp.h>
#include <pwd.h>

/* global signal table */
Import Symbol *_procSignalTable;

Local Value arg, arg2, arg3;
Local int doit(), print_value();
Local void illfmt();
Local char *ctor();
Local Value get_value();

#ifndef MAXPATHLEN
#define MAXPATHLEN	1024	/* take a guess */
#endif

/*********************************
 *   AWL "internal" utilities    *
 * (i.e. not UNIX counterparts)  *
 *********************************/

/*
 * _argc():
 * Returns number of args passed to function.
 */
DEFUN(_argc)
{
     value_set(arg, DATA, INT, int, value_int(awl_stack(aw)[awl_fp(aw) - 1]));
     return(arg);
}

/*
 * _argv(INT):
 * Returns a specific arg number. This is VERY awl specific (I.E. if you
 * change awl's basic frame layout, this will break).
 */
DEFUN(_argv)
{
     arg = get_arg(aw, 1, INT, TRUE);
     if (value_type(arg)) {
	  int old_nparms = value_int(awl_stack(aw)[awl_fp(aw) - 1]);
	  debug(aw, "_argv: fetching arg #%d from frame.", value_int(arg));
	  if (value_int(arg) > old_nparms) {
	       exec_warn(aw, "_argv: Requested arg #%d > %d passed",
			 value_int(arg), old_nparms);
	       value_clear(arg);
	  }
	  else {
	       /*
		* We can't simply use FRAME_OFFSET here because we're
		* loading the value from the previous stack frame, not the
		* current one.
		*/
	       arg = do_load(aw,
			     awl_stack(aw)[value_int(awl_stack(aw)[awl_fp(aw)])
					   - ((old_nparms
					       - value_int(arg))
					      + FRAME_REG_SIZE)],
			     TRUE);
	  }
     }
     return(arg);
}

/*
 * assign(FDESC, NAME):
 * Redirect stdin, stdout or stderr.
 */
DEFUN(assign)
{
     arg = get_arg(aw, 1, ANY, FALSE);
     if (is_filedesc_type(value_type(arg))) {
	  FILE *ifp = (value_type(arg) != FILEPDESC ? value_file(arg)
		       : proc_in(value_any(arg))),
	       *ofp = (value_type(arg) != FILEPDESC ? value_file(arg)
		       : proc_out(value_any(arg)));

	  arg2 = get_arg(aw, 2, STRING, TRUE);
	  if (!stricomp(value_string(arg2), "stdin"))
	       awl_in(aw) = ifp;
	  else if (!stricomp(value_string(arg2), "stdout"))
	       awl_out(aw) = ofp;
	  else if (!stricomp(value_string(arg2), "stderr"))
	       awl_err(aw) = ofp;
	  else
	       exec_warn("assign: Expected stdin, stdout or stderr. Got: %s",
			 value_string(arg2));
     }
     else {
	  exec_warn("assign: Expected type file descriptor, got '%s'",
		    type_name(value_type(arg)));
	  value_clear(arg);
     }
     return(arg);
}

/*
 * classid(name):
 * Returns class id (as int) for name.
 */
DEFUN(classid)
{
     String name;

     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  name = value_string(arg);
	  value_set(arg, DATA, INT, int, map_resword_to_token(name, T_CLASS));
     }
     return(arg);
}

/*
 * classof(value):
 * Returns class of value as an int.
 */
DEFUN(classof)
{
     int tmp;

     arg = get_arg(aw, 1, ANY, FALSE);
     tmp = value_class(arg);
     value_set(arg, DATA, INT, int, tmp);
     return(arg);
}

/*
 * get(type):
 * Get formatted input of "type" from input. If "type" is ANY, try to
 * intuit type.
 */
DEFUN(get)
{
     arg = get_arg(aw, 1, INT, TRUE);
     arg2 = get_value(aw, arg);
     return(arg2);
}

/*
 * input(value, type):
 * Prints "value" to stdout and waits for data of type "type" on stdin.
 * Basically a convenience function for the often used print()/get()
 * pair.
 */
DEFUN(input)
{
     print_value(aw, get_arg(aw, 1, ANY, FALSE));
     return(get_value(aw, get_arg(aw, 2, INT, TRUE)));
}

/*
 * length(type):
 * Returns the length of a string, list or file (coercing expression to
 * a string if necessary).
 */
DEFUN(length)
{
     arg = get_arg(aw, 1, ANY, FALSE);
     if (value_type(arg) == STRING)
	  value_set(arg, DATA, INT, int, strlen(value_string(arg)));
     else if (value_type(arg) == LIST)
	  value_set(arg, DATA, INT, int, list_len(value_list(arg)));
     else if (is_filedesc_type(value_type(arg))) {
	  struct stat sb;
	  FILE *fp;
	  Import int fstat();
	  if (value_type(arg) == FILEDESC)
	       fp = value_file(arg);
	  else
	       fp = (proc_in(value_any(arg)) ? proc_in(value_any(arg))
		     : proc_out(value_any(arg)));

	  fstat(fileno(fp), &sb);
	  value_set(arg, DATA, INT, int, sb.st_size);
     }
     else {
	  Value x;

	  x = coerce(aw, arg, STRING);
	  value_set(arg, DATA, INT, int, strlen(value_string(x)));
     }
     return(arg);
}

/*
 * load(STRING):
 * Load awl code from file STRING. Return 0 if successful, 1 if failure.
 */
DEFUN(load)
{
     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_any(arg)) {
	  String file = value_string(arg);

	  value_set(arg, DATA, INT, int, parseFile(aw, file));
     }
     return(arg);
}

/*
 * print(n...n1):
 * Print values from the stack. This is a good example of a routine
 * handling a variable number of arguments.
 */
DEFUN(print)
{
     int i, len = 0;

     for (i = 1; i <= awl_nparms(aw); i++) {
	  arg = get_arg(aw, i, ANY, FALSE);
	  len += print_value(aw, arg);
     }
     fflush(awl_out(aw));
     value_set(arg, DATA, INT, int, len);
     return(arg);
}

/*
 * println(n..n1):
 * same as print but with a newline automatically appended.
 */
DEFUN(println)
{
     arg = awl_print(aw);
     fputc('\n', awl_out(aw));
     ++value_int(arg);
     return(arg);
}

/*
 * printname(INT):
 * Returns printable version of type/class INT
 */
DEFUN(printname)
{
     String result;

     arg = get_arg(aw, 1, INT, FALSE);
     result = map_token_to_resword(value_int(arg), T_ALL);
     if (result)
	  value_set(arg, DATA, STRING, aobj, new_aobj(XtNewString(result)));
     else
	  value_set(arg, DATA, INT, int, 0);
     return(arg);
}

/*
 * setbuf(FILE, STRING):
 * Set the buffer for a file descriptor.
 */
DEFUN(setbuf)
{
     arg = get_arg(aw, 1, ANY, FALSE);
     if (value_type(arg) == FILEDESC || value_type(arg) == FILEPDESC) {
	  arg2 = get_arg(aw, 2, ANY, FALSE);
	  if (value_type(arg2) == STRING || value_type(arg2) == INT
	      || value_type(arg2) == ANY) {
	       Generic buf1 = (value_type(arg2) == STRING
			       ? (Generic)value_string(arg2)
			       : value_any(arg2));

	       if (value_type(arg) == FILEDESC)
		    setbuf(value_file(arg), buf1);
	       else {
		    if (proc_in(value_any(arg)) && proc_out(value_any(arg))) {
			 arg3 = get_arg(aw, 3, ANY, FALSE);
			 if (value_type(arg3) == STRING
			     || value_type(arg3) == INT
			     || value_type(arg3) == ANY) {
			      Generic buf2 = (value_type(arg3) == STRING
					      ? (Generic)value_string(arg3)
					      : value_any(arg3));
			      setbuf(proc_in(value_any(arg)), buf1);
			      setbuf(proc_out(value_any(arg)), buf2);
			 }
			 else
			      exec_error(aw, "Need 2 buffers for this process descriptor");
		    }
		    else if (proc_in(value_any(arg)))
			 setbuf(proc_in(value_any(arg)), buf1);
		    else if (proc_out(value_any(arg)))
			 setbuf(proc_out(value_any(arg)), buf1);
		    else
			 exec_error(aw, "Process has no open descriptors!");
	       }
	  }
	  else
	       exec_error(aw, "Illegal buffer argument for setbuf");
     }
     else
	  exec_error(aw, "Expected file descriptor, got '%s'",
		     type_name(value_type(arg)));
     return(arg);
}

/*
 * symbol(STRING):
 * Look for a symbol named STRING and return it as-is.
 */
DEFUN(symbol)
{
     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_any(arg)) {
	  String s = value_string(arg);
	  value_set(arg, DATA, SYMBOL, symbol, symbol_find(aw, s));
     }
     return(arg);
}

/*
 * typeid(name):
 * Returns type id (as int) for name.
 */
DEFUN(typeid)
{
     String name;

     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  name = value_string(arg);
	  value_set(arg, DATA, INT, int, map_resword_to_token(name, T_TYPE));
     }
     return(arg);
}

/*
 * typeof(value):
 * Returns type of value as an int.
 */
DEFUN(typeof)
{
     int tmp;

     arg = get_arg(aw, 1, ANY, FALSE);
     tmp = value_type(arg);
     value_set(arg, DATA, INT, int, tmp);
     return(arg);
}


/************************************
 *    UNIX interface functions      *
 ************************************/

/*
 * chdir(STRING):
 * Do a chdir(STRING).
 */
DEFUN(chdir)
{
     Import int chdir();
     int status = -1;

     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg))
	  status = chdir(value_string(arg));
     value_set(arg, DATA, INT, int, status);
     return(arg);
}

/*
 * chmod(STRING, INT):
 * chmod() file STRING to mode INT.
 */
DEFUN(chmod)
{
     int status = -1;
     Import int chmod();

     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  arg2 = get_arg(aw, 2, INT, TRUE);
	  if (value_type(arg2))
	       status = chmod(value_string(arg), value_int(arg2));
     }
     value_set(arg, DATA, INT, int, status);
     return(arg);
}

/*
 * chown(STRING, INT1, INT2):
 * chown() file STRING to user INT1 and group INT2.
 */
DEFUN(chown)
{
     int status = -1;
     Import int chown(), fchown();

     arg = get_arg(aw, 1, ANY, FALSE);
     if (value_type(arg) == STRING || value_type(arg) == FILEDESC) {
	  arg2 = get_arg(aw, 2, INT, TRUE);
	  if (value_type(arg2)) {
	       arg3 = get_arg(aw, 3, INT, TRUE);
	       if (value_type(arg3)) {
		    if (value_type(arg) == STRING)
			 status = chown(value_string(arg),
					value_int(arg2),
					value_int(arg3));
		    else
			 status = fchown(fileno(value_file(arg)),
					 value_int(arg2),
					 value_int(arg3));
	       }
	  }
     }
     else {
	  exec_warn(aw, "chown: Expected file name or descriptor, got '%s'",
		    type_name(value_type(arg)));
     }
     value_set(arg, DATA, INT, int, status);
     return(arg);
}

/*
 * chroot(STRING):
 * Change root directory to STRING.
 */
DEFUN(chroot)
{
     int status = -1;
     Import int chroot();

     arg = get_arg(aw, 1, STRING, TRUE);
     if (type_name(arg)) {
	  status = chroot(value_string(arg));
     }
     value_set(arg, DATA, INT, int, status);
     return(arg);
}

/*
 * close(FDESC):
 * close() a file descriptor or process.
 */
DEFUN(close)
{
     int status = EOF;

     arg = get_arg(aw, 1, ANY, FALSE);
     if (value_type(arg) == FILEDESC)
	  status = fclose(value_file(arg));
     else if (value_type(arg) == FILEPDESC)
	  status = p_pclose(aw, value_any(arg));
     else
	  exec_warn("close: Expected file descriptor, got '%s'",
		    type_name(value_type(arg)));
     value_set(arg, DATA, INT, int, status);
     return(arg);
}

/*
 * exec(STRING, LIST):
 * execvp() file STRING with arguments LIST.
 */
DEFUN(exec)
{
     Import int execvp();

     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  arg2 = get_arg(aw, 2, LIST, TRUE);
	  if (value_type(arg2)) {
	       execvp(value_string(arg), value_list(arg2));
	       value_set(arg, DATA, INT, int, -1);	/* we failed */
	  }
     }
     return(arg); /* maybe */
}

/*
 * exit(INT):
 * exit() with return code (INT)expr.
 */
DEFUN(exit)
{
     arg = get_arg(aw, 1, INT, FALSE);
     exit(value_int(arg));
     return(arg); /* sure hope not! */
}

/*
 * fork():
 * fork() a process. Wheee... This may get you into trouble
 * if you don't handle your file descriptors properly.
 */
DEFUN(fork)
{
     Import int fork();

     value_set(arg, DATA, INT, int, fork());
     return(arg);
}

/*
 * free(value):
 * Free's previously malloc'd value.
 */
DEFUN(free)
{
     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  if (value_class(arg) != ALLOCATED)
	       exec_warn(aw, "attempt to free non-malloc'd variable");
	  else {
	       if (--aobj_cnt(value_aobj(arg)) <= 0) {
		    XtFree(value_string(arg));
		    XtFree(value_aobj(arg));
		    value_clear(arg);
	       }
	  }
     }
     return(arg);
}

/*
 * fmt_time(fmt, time):
 * Format time description according to fmt.
 */
DEFUN(fmt_time)
{
     String res;

     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  arg2 = get_arg(aw, 2, ANY, FALSE);
	  if ((res = format_time(value_string(arg), value_any(arg2))) != NULL)
	       value_set(arg, DATA, STRING, aobj, new_aobj(res));
	  else
	       value_clear(arg);
     }
     return(arg);
}

/*
 * getenv(STRING):
 * Returns environment variable STRING.
 */
DEFUN(getenv)
{
     String cp = NULL;
     Import String getenv();

     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  if ((cp = getenv(value_string(arg))) != NULL)
	       cp = XtNewString(cp);
     }
     if (cp)
	  value_set(arg, DATA, STRING, aobj, new_aobj(cp));
     else
	  value_clear(arg);
     return(arg);
}

/*
 * getegid():
 * Return our effective GID.
 */
DEFUN(getegid)
{
     Import int getegid();

     value_set(arg, DATA, INT, int, getegid());
     return(arg);
}

/*
 * geteuid():
 * Return our effective UID.
 */
DEFUN(geteuid)
{
     Import int geteuid();

     value_set(arg, DATA, INT, int, geteuid());
     return(arg);
}

/*
 * getgid():
 * Return our GID.
 */
DEFUN(getgid)
{
     Import int getgid();

     value_set(arg, DATA, INT, int, getgid());
     return(arg);
}

/*
 * getpid()
 * Return our PID.
 */
DEFUN(getpid)
{
     Import int getpid();

     value_set(arg, DATA, INT, int, getpid());
     return(arg);
}

/*
 * getuid():
 * Return our UID.
 */
DEFUN(getuid)
{
     Import int getuid();

     value_set(arg, DATA, INT, int, getuid());
     return(arg);
}

/*
 * getwd():
 * return the current working directory.
 */
DEFUN(getwd)
{
     Import String getwd();
     char name[MAXPATHLEN + 1];
     String cp = NULL;

     if (getwd(name))
	  cp = XtNewString(name);
     if (cp)
	  value_set(arg, DATA, STRING, aobj, new_aobj(cp));
     else
	  value_clear(arg);
     return(arg);
}
     
/*
 * group(ANY):
 * Return the group name for group ID or group ID for group name.
 */
DEFUN(group)
{
     Import struct group *getgrnam(), *getgrgid();
     struct group *id = NULL;
     int typ;

     arg = get_arg(aw, 1, ANY, FALSE);
     typ = value_type(arg);

     if (typ == STRING)
	  id = getgrnam(value_string(arg));
     else if (typ == INT) {
	  int x = value_int(arg);

	  if (x < 0)
	       x = -x;
	  id = getgrgid(x);
     }
     if (id) {
	  if (typ == STRING)
	       value_set(arg, DATA, INT, int, id->gr_gid);
	  else if (typ == INT) {
	       if (value_int(arg) < 0)
		    value_set(arg, DATA, LIST, aobj,
			      new_aobj(list_dup(aw, id->gr_mem)));
	       else
		    value_set(arg, DATA, STRING, aobj,
			      new_aobj(XtNewString(id->gr_name)));
	  }
     }
     return(arg);
}

/*
 * kill(proc, sig):
 * Try to send sig to pid/process file desc.
 */
DEFUN(kill)
{
     Import int kill();

     arg = get_arg(aw, 1, ANY, FALSE);
     if (value_type(arg) == INT || value_type(arg) == FILEPDESC) {
	  arg2 = get_arg(aw, 2, INT, TRUE);
	  if (value_type(arg2)) {
	       int pid = (value_type(arg) == INT ? value_int(arg)
			  : proc_pid(value_any(arg)));

	       value_set(arg, DATA, INT, int, kill(pid, value_int(arg2)));
	  }
     }
     else
	  exec_warn(aw, "kill: expected pid or process file desc. Got '%s'",
		    type_name(value_type(arg)));
     return(arg);
}

/*
 * malloc(INT):
 * Returns INT bytes of allocated space.
 */
DEFUN(malloc)
{
     String cp;
     int size;

     arg = get_arg(aw, 1, INT, TRUE);
     if (value_type(arg)) {
	  size = value_int(arg);
	  if ((cp = XtMalloc(size)) != NULL) {
	       bzero(cp, size);
	       value_set(arg, ALLOCATED, STRING, aobj, new_aobj(cp));
	  }
	  else
	       value_clear(arg);
     }
     return(arg);
}

/*
 * open(NAME, MODE):
 * open() a file or process with a given mode.
 */
DEFUN(open)
{
     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  arg2 = get_arg(aw, 2, STRING, TRUE);
	  if (value_type(arg2)) {
	       FILE *fp;
	       int type;
	       char mode[3];

	       strncpy(mode, value_string(arg2), 2);
	       mode[2] = '\0';
	       if (mode[1] == 'p') {
		    mode[1] = '\0';
		    value_set(arg, DATA, FILEPDESC, any,
			      p_popen(aw, value_string(arg), mode));
	       }
	       else {
		    value_set(arg, DATA, FILEDESC, file,
			      fopen(value_string(arg), mode));
	       }
	  }
     }
     return(arg);
}

/*
 * perror(STRING):
 * Call perror(STRING).
 */
DEFUN(perror)
{
     Import void perror();

     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  perror(value_string(arg));
     }
     return(arg);
}

/*
 * printf(format [,arg] ...)
 * Do a printf() using our own sprintf() call.
 */
DEFUN(printf)
{
     int cnt;
     Import Value awl_sprintf();

     arg = awl_sprintf(aw);
     cnt = fprintf(awl_out(aw), value_string(arg));
     do_free(&arg);
     value_set(arg, DATA, INT, int, cnt);
     return(arg);
}

/*
 * read(dest, cnt):
 * read cnt bytes from stdin, storing in dest.
 */
DEFUN(read)
{
     int result;

     arg = get_arg(aw, 1, ANY, FALSE);
     arg2 = get_arg(aw, 2, INT, TRUE);
     if (value_type(arg) == STRING)
	  result = fread(value_string(arg), 1, value_int(arg2), awl_in(aw));
     else
	  result = fread(value_any(arg), 1, value_int(arg2), awl_in(aw));
     value_set(arg, DATA, INT, int, result);
     return(arg);
}

/*
 * readdir(STRING):
 * Slurp in directory STRING and return list of files therein.
 * (note: This does *not* match the behaviour of its unix namesake, but
 * I like this approach better).
 */

#define LIST_HUNK	32

DEFUN(readdir)
{
     DIR *frobozz;
     struct dirent *dp;
     int lsize, lmax;
     String *ll;

     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  frobozz = opendir(value_string(arg));
	  if (frobozz) {
	       lsize = lmax = 0;
	       while ((dp = readdir(frobozz)) != NULL) {
		    if (!lmax) {
			 lmax = LIST_HUNK;
			 ll = (String *)XtMalloc(lmax * sizeof(String *));
		    }
		    else if (lsize == lmax) {
			 lmax += LIST_HUNK;
			 ll = (String *)XtRealloc(ll, lmax * sizeof(String *));
		    }
		    ll[lsize++] = XtNewString(dp->d_name);
	       }
	       ll[lsize] = NULL;
	       closedir(frobozz);
	       value_set(arg, DATA, LIST, aobj, new_aobj(ll));
	  }
	  else
	       value_clear(arg);
     }
     else
	  value_clear(arg);
     return(arg);
}

/*
 * setgid(INT):
 * Do a setgid() call.
 */
DEFUN(setgid)
{
     Import int setgid();

     arg = get_arg(aw, 1, INT, TRUE);
     if (value_type(arg)) {
	  int gid = value_int(arg);

	  value_set(arg, DATA, INT, int, setgid(gid));
     }
     return(arg);
}

/*
 * setuid(INT):
 * Do a setuid() call.
 */
DEFUN(setuid)
{
     Import int setuid();

     arg = get_arg(aw, 1, INT, TRUE);
     if (value_type(arg)) {
	  int uid = value_int(arg);

	  value_set(arg, DATA, INT, int, setuid(uid));
     }
     return(arg);
}

/*
 * signal(INT, sym):
 * Set signal routine for a given signal INT.
 */
DEFUN(signal)
{
     Symbol oldval = (Symbol)NULL;

     arg = get_arg(aw, 1, INT, TRUE);
     if (value_type(arg)) {
	  int signo = value_int(arg);

	  if (signo < 0 || signo >= MAXSIG)
	       exec_warn(aw, "Illegal signal number %d", value_int(arg));
	  else {
	       arg2 = get_arg(aw, 2, ANY, FALSE);
	       if (!_procSignalTable) {
		    _procSignalTable =
			 (Symbol *)XtMalloc(sizeof(Symbol) * MAXSIG);
		    bzero(_procSignalTable, sizeof(Symbol) * MAXSIG);
	       }
	       oldval = _procSignalTable[signo];
	       /*
		* Since signals are process global and there is no limit on
		* the number of awl widgets a process may have, we warn the
		* user of any conflicts between awl widget instances using
		* the signal mechanism.
		*/
	       if (oldval > (Symbol)MY_SIG_IGN && aw != symbol_root(oldval))
		    exec_warn(aw, "Widget %x usurping handler for signal %d",
			      signo);
	       if (value_type(arg2) == INT) {
		    if (value_int(arg2) == MY_SIG_IGN)
			 signal(signo, SIG_IGN);
		    else if (value_int(arg2) == MY_SIG_DFL)
			 signal(signo, SIG_DFL);
		    else
			 exec_error(aw, "Illegal int passed as signal func");
		    _procSignalTable[signo] = (Symbol)value_int(arg2);
	       }
	       else if (value_type(arg2) == SYMBOL) {
		    if (symbol_is_function(value_symbol(arg2))) {
			 _procSignalTable[signo] = value_symbol(arg2);
			 signal(signo, do_signal_user);
		    }
		    else
			 exec_error(aw, "Signal handler is not a function");
	       }
	       else
		    exec_error(aw, "Signal handler must be a function symbol");
	  }
     }
     value_set(arg, DATA, SYMBOL, symbol, oldval);
     return(arg);
}

/*
 * sleep(INT):
 * sleep() for INT seconds.
 */
DEFUN(sleep)
{
     Import unsigned int sleep();

     arg = get_arg(aw, 1, INT, TRUE);
     if (value_type(arg))
	  sleep(value_int(arg));
     return(arg);
}

/*
 * sprintf(var, format [,arg] ...)
 * Do a sprintf to "var".
 *
 * This routine was taken almost verbatim from Chris Torek and Fred Blonder's
 * Uncopyrighted "shell printf()" code. See the man page.
 *
 */
DEFUN(sprintf)
{
     register String cp, convp, targstr;
     register int ch, ndyn, flags;
     char cbuf[BUFSIZ];	/* separates each conversion */
     Local char hasmod[] = "has integer length modifier";
     Value target;
     int tindex, aa, nargs, tsize;

     /* flags */
#define	LONG	1
#define	SHORT	2

     nargs = awl_nparms(aw);
     if (nargs < 1) {
	  value_set(arg, DATA, INT, int, 0);
	  return(arg);
     }
     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  /* try to guess how much space we'll need */
	  tsize = strlen(value_string(arg));
	  for (aa = 1; aa <= nargs; aa++) {
	       target = get_arg(aw, aa, ANY, FALSE);
	       if (value_type(target) == STRING && value_string(target))
		    tsize += strlen(value_string(target));
	       else /* guesstimate generously */
		    tsize += 28;
	  }
	  targstr = XtMalloc(tsize);
	  bzero(targstr, tsize);
	  value_set(target, DATA, STRING, aobj, new_aobj(targstr));

	  aa = 1;
	  tindex = 0;
	  cp = value_string(arg);

	  /*
	   * Scan format string for conversion specifications.
	   * (The labels would be loops, but then everything falls
	   * off the right.)
	   */
scan:
	  while ((ch = *(cp++)) != '%') {
	       if (ch == '\0')
		    return(target);
	       targstr[tindex++] = ch;
	  }

	  ++aa;
	  ndyn = 0;
	  flags = 0;
	  convp = cbuf;
	  *(convp++) = ch;

	  /* scan for conversion character */
cvt:
	  switch (ch = *(cp++)) {

	  case '\0':	/* unterminated conversion */
	       targstr[tindex] = '\0';
	       return(target);

	       /* string or character format */
	  case 'c': case 's':
	       if (flags) {
		    illfmt(aw, cbuf, convp, ch, hasmod);
		    targstr[tindex] = '\0';
		    return(target);
	       }
	       if (!doit(aw, cbuf, convp, aa, nargs, targstr, &tindex,
			 ndyn, ch, ch)) {
		    targstr[tindex] = '\0';
		    return(target);
	       }
	       goto scan;
	       
	       /* integer formats */
	  case 'd': case 'i': case 'o': case 'u': case 'x': case 'X':
	       if ((flags & (LONG|SHORT)) == (LONG|SHORT)) {
		    illfmt(aw, cbuf, convp, ch, "is both long and short");
		    targstr[tindex] = '\0';
		    return(target);
	       }
	       if (!doit(aw, cbuf, convp, aa, nargs, targstr, &tindex,
			 ndyn, ch, (flags & LONG ? 'l' : flags & SHORT ? 'h' :
			 'i'))) {
		    targstr[tindex] = '\0';
		    return(target);
	       }
	       goto scan;

	       /* floating point formats */
	  case 'e': case 'E': case 'f': case 'g': case 'G':
	       if (flags) {
		    illfmt(aw, cbuf, convp, ch, hasmod);
		    targstr[tindex] = '\0';
		    return(target);
	       }
	       if (!doit(aw, cbuf, convp, aa, nargs, targstr, &tindex,
			 ndyn, ch, 'f')) {
		    targstr[tindex] = '\0';
		    return(target);
	       }
	       goto scan;

	       /* Roman (well, why not?) */
	  case 'r': case 'R':
	       if (flags) {
		    illfmt(aw, cbuf, convp, ch, hasmod);
		    targstr[tindex] = '\0';
		    return(target);
	       }
	       if (!doit(aw, cbuf, convp, aa, nargs, targstr,  &tindex,
			 ndyn, 's', ch)) {
		    targstr[tindex] = '\0';
		    return(target);
	       }
	       goto scan;

	  case '%':	/* boring */
	       targstr[tindex++] = '%';
	       goto scan;

	       /* short integers */
	  case 'h':
	       flags |= SHORT;
	       break;

	       /* long integers */
	  case 'l':
	       flags |= LONG;
	       break;

	       /* field-width or precision specifier, or flag: keep scanning */
	  case '.': case '#': case '-': case '+': case ' ':
	  case '0': case '1': case '2': case '3': case '4':
	  case '5': case '6': case '7': case '8': case '9':
	       break;

	       /* dynamic field width or precision: count it */
	  case '*':
	       ndyn++;
	       break;

	  default:	/* something we cannot handle */
	       if (isascii(ch) && isprint(ch))
		    cbuf[0] = ch, cbuf[1] = 0;
	       else
		    (void)sprintf(cbuf, "\\%03o", (unsigned char)ch);
	       exec_warn(aw, "vprintf: illegal conversion character `%s'",
			 cbuf);
	       targstr[tindex] = '\0';
	       return(target);
	       /* NOTREACHED */
	  }

	  /* 2 leaves room for ultimate conversion char and for \0 */
	  if (convp >= &cbuf[sizeof(cbuf) - 2]) {
	       exec_warn(aw, "vprintf: conversion string too long");
	       targstr[tindex] = '\0';
	       return(target);
	  }
	  *(convp++) = ch;
	  goto cvt;
     }
     return(target);
}

/*
 * stat(FMT, FILE):
 * Stat a file name or file descriptor, returning stat information according
 * to format string.
 */
DEFUN(stat)
{
     struct stat sb;
     Value ret;
     Import int stat(), fstat();

     value_clear(ret);
     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  arg2 = get_arg(aw, 2, ANY, FALSE);
	  if (value_type(arg2) == STRING) {
	       if (!stat(value_string(arg2), &sb))
		    value_set(ret, DATA, STRING, aobj,
			      new_aobj(stat_expand(aw, value_string(arg2),
						   value_string(arg), &sb)));
	  }
	  else if (is_filedesc_type(value_type(arg2))) {
	       FILE *fp;
	       char name[80];

	       fp = (value_type(arg2) == FILEDESC ? value_file(arg2)
		     : (proc_in(value_any(arg2)) ? proc_in(value_any(arg2))
			: proc_out(value_any(arg2))));

	       if (!fstat(fileno(fp), &sb)) {
		    sprintf(name, "<fdesc#%d>", value_file(arg2));
		    value_set(ret, DATA, STRING, aobj,
			      new_aobj(stat_expand(aw, name,
						   value_string(arg), &sb)));
	       }
	  }
	  else
	       exec_warn(aw, "stat: Expected file name or descriptor, got %s",
			 type_name(value_type(arg2)));
     }
     return(ret);
}

/*
 * system(STRING):
 * Do a system(STRING) call.
 */
DEFUN(system)
{
     arg = get_arg(aw, 1, STRING, TRUE);
     if (value_type(arg)) {
	  int status = system(value_string(arg));
	  value_set(arg, DATA, INT, int, status);
     }
     return(arg);
}

/*
 * time():
 * Return the time in seconds since 00:00:00 GMT, January 1, 1970.
 */
DEFUN(time)
{
     time_t ret;
     Import time_t time();

     ret = time(0);
     value_set(arg, DATA, ANY, any, (Generic)ret);
     return(arg);
}

/*
 * user(ANY):
 * Return user name for user ID or user ID for user name.
 */
DEFUN(user)
{
     int typ;
     Import struct passwd *getpwnam(), *getpwuid();
     struct passwd *pw = NULL;

     arg = get_arg(aw, 1, ANY, FALSE);
     typ = value_type(arg);

     if (typ == STRING)
	  pw = getpwnam(value_string(arg));
     else if (typ == INT)
	  pw = getpwuid(value_int(arg));
     if (pw) {
	  if (typ == STRING)
	       value_set(arg, DATA, INT, int, pw->pw_uid);
	  else if (typ == INT)
	       value_set(arg, DATA, STRING, aobj,
			 new_aobj(XtNewString(pw->pw_name)));
     }
     return(arg);
}
     
/*
 * write(dest, cnt):
 * write cnt bytes from dest to stdout.
 */
DEFUN(write)
{
     int result;

     arg = get_arg(aw, 1, ANY, FALSE);
     arg2 = get_arg(aw, 2, INT, TRUE);
     if (value_type(arg) == STRING)
	  result = fwrite(value_string(arg), 1, value_int(arg2), awl_out(aw));
     else
	  result = fwrite(value_any(arg), 1, value_int(arg2), awl_out(aw));
     value_set(arg, DATA, INT, int, result);
     return(arg);
}


/******************************************
 *  Various local work functions needed   *
 *  by the systems interface              *
 ******************************************/

/*
 * These next two functions are for sprintf/printf.
 */
Local void illfmt(aw, cbuf, convp, ch, why)
AwlWidget aw;
char *cbuf, *convp;
int ch;
char *why;
{

     *(convp++) = ch;
     *convp = 0;
     exec_warn(aw, "vprintf: format `%s' illegal: %s", cbuf, why);
}

/*
 * Emit a conversion.  cch holds the printf format character for
 * this conversion; cty holds a simplified version (all integer
 * conversions, e.g., are represented as 'i').
 */
Local int doit(aw, cbuf, convp, argn, nargs, targ, tidx, ndyn, cch, cty)
AwlWidget aw;
char *cbuf, *convp;
int argn, nargs;
String targ;
int *tidx, ndyn;
int cch, cty;
{
     char *s;
     union {		/* four basic conversion types */
	  int i;
	  long l;
	  FTYPE d;
	  String str;
     } arg;
     int a1, a2;	/* dynamic width and/or precision */

     /* finish off the conversion string */
     s = convp;
     *(s++) = cch;
     *s = 0;
     s = cbuf;

     /* verify number of arguments */
     if (argn > nargs) {
	  exec_warn(aw, "vprintf: not enough args for format `%s'", s);
	  return(0);
     }

     /* pick up dynamic specifiers */
     if (ndyn) {
	  a1 = value_int(get_arg(aw, argn, INT, FALSE));
	  if (ndyn > 1)
	       a2 = value_int(get_arg(aw, argn, INT, FALSE));
	  if (ndyn > 2) {
	       exec_warn(aw, "vprintf: too many `*'s in `%s'", s);
	       return(0);
	  }
     }

#define	PRINTF(what) \
     if (ndyn == 0) \
	  *tidx += sprintf(targ + *tidx, s, what); \
     else if (ndyn == 1) \
	  *tidx += sprintf(targ + *tidx, s, a1, what); \
     else \
	  *tidx += sprintf(targ + *tidx, s, a1, a2, what);

     /* emit the appropriate conversion */
     switch (cty) {

	  /* string */
     case 's':
	  arg.str = value_string(get_arg(aw, argn, STRING, FALSE));
	  if (arg.str)
	       backslash_eliminate(arg.str, NORMAL_ELIMINATE, 0);
	  else
	       arg.str = "(null)";
	  goto string;

	  /* roman (much like string) */
     case 'r': case 'R':
	  arg.str = ctor(value_int(get_arg(aw, argn, INT, FALSE)), cty == 'R');
	  goto string;

     string:
	  PRINTF(arg.str);
	  break;

	  /* floating point */
     case 'f':
	  arg.d = value_float(get_arg(aw, argn, FLOAT, FALSE));
	  PRINTF(arg.d);
	  break;

	  /* character */
     case 'c':
	  arg.i = (int)value_char(get_arg(aw, argn, CHAR, FALSE));
	  goto integer;

	  /* short integer */
     case 'h':
	  arg.i = (short)value_int(get_arg(aw, argn, INT, FALSE));
	  goto integer;

	  /* integer */
     case 'i':
	  arg.i = value_int(get_arg(aw, argn, INT, FALSE));
	  goto integer;

     integer:
	  PRINTF(arg.i);
	  break;

	  /* long integer */
     case 'l':
	  arg.l = (long)value_int(get_arg(aw, argn, INT, FALSE));
	  PRINTF(arg.l);
	  break;
     }
     return (1);
}
/*
 * Convert integer to Roman Numerals. (How have you survived without it?)
 * [That's a very good question, Chris..]
 */
Local char *ctor(x, caps)
int x, caps;
{
     Local char buf[BUFSIZ];
     register char *outp = buf;
     register unsigned n = x;
     register int u, v;
     register char *p, *q;

     if ((int)n < 0) {
	  *(outp++) = '-';
	  n = -n;
     }
     p = caps ? "M\2D\5C\2L\5X\2V\5I" : "m\2d\5c\2l\5x\2v\5i";
     v = 1000;
     if (n >= v * BUFSIZ / 2)	/* conservative */
	  return ("[abortive Roman numeral]");
     for (;;) {
	  while (n >= v)
	       *(outp++) = *p, n -= v;
	  if (n == 0)
	       break;
	  q = p + 1;
	  u = v / *q;
	  if (*q == 2)		/* magic */
	       u /= *(q += 2);
	  if (n + u >= v) {
	       *(outp++) = *(++q);
	       n += u;
	  } else {
	       p++;
	       v /= *(p++);
	  }
     }
     *outp = 0;
     return (buf);
}

/* For get() and input() */
Local Value get_value(aw, arg)
AwlWidget aw;
Value arg;
{
     register char ch;
     String inbuf = NULL;
     int i = 0, max = 0;
     Value ret;
     Import Boolean isValidDigit();
#ifndef toupper
     Import char toupper();
#endif

     switch (value_int(arg)) {
     case ANY:
	  while ((ch = fgetc(awl_in(aw))) != EOF && ch != awl_sep(aw)[0])
	       append_string(&inbuf, &max, i++, ch);
	  append_string(&inbuf, &max, i, '\0');
	  if (isValidDigit(aw, *inbuf, ANY)) {
	       if (index(inbuf, '.'))
		    value_set(ret, DATA, FLOAT, float, atof(inbuf));
	       else
		    value_set(ret, DATA, INT, int, atonum(aw, inbuf, 0));
	       XtFree(inbuf);
	  }
	  else {
	       if (i || ch == awl_sep(aw)[0])
		    value_set(ret, DATA, STRING, aobj, new_aobj(inbuf));
	       else {
		    value_set(ret, DATA, INT, int, 0);
		    XtFree(inbuf);
	       }
	  }
	  break;

     case INT:
	  while ((ch = fgetc(awl_in(aw))) != EOF) {
	       if (isspace(ch) && !i)
		    continue;
	       else if (isValidDigit(aw, ch, INT))
		    append_string(&inbuf, &max, i++, ch);
	       else if (ch == '.')
		    exec_warn(aw, "decimal point ignored for INT get/input");
	       else
		    break;
	  }
	  append_string(&inbuf, &max, i, '\0');
	  value_set(ret, DATA, INT, int, atonum(aw, inbuf, 0));
	  XtFree(inbuf);
	  break;

     case CHAR:
	  value_set(ret, DATA, CHAR, char, fgetc(awl_in(aw)));
	  break;

     case FLOAT:
	  while ((ch = fgetc(awl_in(aw))) != EOF) {
	       if (isspace(ch) && !i)
		    continue;
	       else if (isValidDigit(aw, ch, FLOAT))
		    append_string(&inbuf, &max, i++, ch);
	       else
		    break;
	  }
	  append_string(&inbuf, &max, i, '\0');
	  value_set(ret, DATA, FLOAT, float, atof(inbuf));
	  XtFree(inbuf);
	  break;

     case STRING:
	  while ((ch = fgetc(awl_in(aw))) != EOF && ch != awl_sep(aw)[0])
	       append_string(&inbuf, &max, i++, ch);
	  append_string(&inbuf, &max, i, '\0');
	  if (i || ch == awl_sep(aw)[0])
	       value_set(ret, DATA, STRING, aobj, new_aobj(inbuf));
	  else {
	       value_set(ret, DATA, INT, int, 0);
	       XtFree(inbuf);
	  }
	  break;

     default:
	  exec_error(aw, "Illegal type '%s' for get/input", value_string(arg));
	  break;
     }
     return(ret);
}

/* For print()/println()/input() */
Local int print_value(aw, arg)
AwlWidget aw;
Value arg;
{
     String tmp;
     int ret = 0;

     if (value_type(arg) == LIST) {
	  String *v = value_list(arg);

	  if (!v)
	       fprintf(awl_out(aw), "(null list)");
	  else while (*v) {
	       ret += (strlen(*v) + 1);
	       fprintf(awl_out(aw), "%s", *(v++));
	       fputs(awl_sep(aw), awl_out(aw));
	  }
     }
     else {
	  tmp = value_string_value(aw, arg);
	  ret += strlen(tmp);
	  fprintf(awl_out(aw), "%s", tmp);
     }
     return(ret);
}
