#ifndef lint
static char *rcsid = "$Header: /usr/src/local/awl/RCS/xtkrtns.c,v 2.7 90/04/19 20:06:27 jkh Exp $";
#endif

/*
 *
 *                     Copyright 1989
 *                    Jordan K. Hubbard
 *
 *		  PCS Computer Systeme, GmbH.
 *	             Munich, West Germany
 *
 *
 * This file is part of AWL.
 *
 * AWL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * AWL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AWL; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*
 * All widget-specific built-in functions. This file is included from
 * builtins.c.
 *
 * $Log:	xtkrtns.c,v $
 * Revision 2.7  90/04/19  20:06:27  jkh
 *  Alpha checkin.
 * 
 * Revision 2.6  90/03/29  18:29:27  jkh
 * Even more bug fixes to XgInstallColorTable
 * 
 * Revision 2.5  90/03/27  09:05:09  jkh
 * Fixed typo.
 * 
 * Revision 2.4  90/03/27  08:35:08  jkh
 * Added Yet Another Sanity Check to XgInstallColorTable().
 * 
 * Revision 2.3  90/03/27  08:31:24  jkh
 * Added sanity check to XgInstallColorTable() for color indices of
 * bad length. Removed insane sanity check for rectangular list (what was
 * I thinking?).
 * 
 * Revision 2.2  90/03/27  08:15:31  jkh
 * Switched back from XtColormap (too gross) to DefaultColormapOfScreen(),
 * which is marginally less gross.
 * 
 * Revision 2.1  90/03/27  08:02:30  jkh
 * Switched over to XtColormap macro in XgInstallColorTable().
 * 
 * Revision 2.0  90/03/26  01:43:59  jkh
 * pre-beta check-in
 * 
 */

#include "AwlP.h"
#include "y.tab.h"

Local Value arg, arg2;

/* Various local helper macros. Don't look. */

#define GET_WIDGET_OR_LAYOUT(name) \
     Widget tmp; \
     arg = get_arg(aw, 1, ANY, FALSE); \
     if (value_type(arg) == LAYOUT) \
	  tmp = layout_id(aw, value_any(arg)); \
     else if (value_type(arg) == WIDGET) \
	  tmp = value_widget(arg); \
     else \
	  exec_error(aw, "%s: Expected widget or layout, got '%s'", \
		     name, type_name(value_type(arg)))

#if defined(__STDC__) && !defined(UNIXCPP)	/* ANSI */

#define DEFUN_XT_VAL_FUNC(xtcall, valtype, valmember) \
DEFUN(xtcall) { GET_WIDGET_OR_LAYOUT(#xtcall); value_type(arg) = valtype; \
		value_##valmember(arg) = xtcall(tmp); return(arg); }

#define DEFUN_XT_OP(name, xtcall) \
DEFUN(name) { GET_WIDGET_OR_LAYOUT(#name); xtcall(tmp); return(arg); }

#define DEFUN_XT_2OP(xtcall, arg2_type, valmember) \
DEFUN(xtcall) { GET_WIDGET_OR_LAYOUT(#xtcall); \
		arg2 = get_arg(aw, 2, arg2_type, TRUE); \
		xtcall(tmp, value_##valmember(arg2)); return(arg); }

#else /* !ANSI */

#define DEFUN_XT_VAL_FUNC(xtcall, valtype, valmember) \
DEFUN(xtcall) { GET_WIDGET_OR_LAYOUT("xtcall"); value_type(arg) = valtype; \
		value_/**/valmember(arg) = xtcall(tmp); return(arg); }

#define DEFUN_XT_OP(name, xtcall) \
DEFUN(name) { GET_WIDGET_OR_LAYOUT("name"); xtcall(tmp); return(arg); }

#define DEFUN_XT_2OP(xtcall, arg2_type, valmember) \
DEFUN(xtcall) { GET_WIDGET_OR_LAYOUT("xtcall"); \
		arg2 = get_arg(aw, 2, arg2_type, TRUE); \
		xtcall(tmp, value_/**/valmember(arg2)); return(arg); }
#endif

Import void XgListSelectItem(), XgListUnselectItem(), XgListChange();

/* The generic functions */
DEFUN_XT_2OP(XgListSelectItem, INT, int)
DEFUN_XT_2OP(XgListUnselectItem, INT, int)
DEFUN_XT_2OP(XgListChange, LIST, list)

DEFUN(XgInstallColorTable)
{
     arg = get_arg(aw, 1, LIST, TRUE);
     if (value_type(arg)) {
	  register CTable ct;
	  register String *l;
	  register int i, j, p = 0;
	  XColor close, exact;
	  Display *dpy = XtDisplay(aw);

	  l = value_list(arg);
	  i = list_len(l);
	  if (!i || i % 2)
    	       exec_warn(aw, "Got odd-size or NULL list as color table");
	  else {
	       if (ct = awl_ctable(aw)) {
		    hash_purge(ctab_hash(ct), NULL);
		    XFreeColors(XtDisplay(aw),
				DefaultColormapOfScreen(XtScreen(aw)),
				ctab_colors(ct), ctab_ncolors(ct), AllPlanes);
		    XtFree(ctab_colors(ct));
	       }
	       else {
		    ct = awl_ctable(aw) = (CTable)XtMalloc(ctab_size);
		    ctab_hash(ct) = hash_create(ALIST_HASH_SIZE);
	       }
	       if ((ctab_nchars(ct) = strlen(l[0])) >= 31) {
		    exec_warn(aw, "Insane index length for first table entry");
		    return(arg);
	       }
	       ctab_ncolors(ct) = i / 2;
	       ctab_colors(ct) = (Pixel *)XtMalloc(sizeof(Pixel) * (i / 2));
	       for (j = 0; j < i; j += 2) {
		    if (strlen(l[j]) != ctab_nchars(ct)) {
			 exec_warn(aw, "Color tag '%s' is not of length %d",
				   l[j], ctab_nchars(ct));
			 l[j][ctab_nchars(ct)] = '\0';
		    }
		    XAllocNamedColor(XtDisplay(aw),
				     DefaultColormapOfScreen(XtScreen(aw)),
				     l[j + 1], &close, &exact);
		    ctab_colors(ct)[p] = close.pixel;
		    hash_search(ctab_hash(ct), l[j], p++, NULL);
	       }
	       value_set(arg, DATA, INT, int, ctab_ncolors(ct));
	  }
     }
     return(arg);
}

DEFUN(XtGetSelectionTimeout)
{
     value_set(arg, DATA, INT, int,
	       XtAppGetSelectionTimeout(XtWidgetToApplicationContext(aw)));
     return(arg);
}

DEFUN_XT_VAL_FUNC(XtClass, WCLASS, wclass)
DEFUN_XT_VAL_FUNC(XtIsManaged, INT, int)
DEFUN_XT_VAL_FUNC(XtIsRealized, INT, int)
DEFUN_XT_VAL_FUNC(XtIsSensitive, INT, int)
DEFUN_XT_VAL_FUNC(XtParent, WIDGET, widget)
DEFUN_XT_VAL_FUNC(XtSuperclass, WCLASS, wclass)

DEFUN_XT_OP(XtRealize, XtRealizeWidget)
DEFUN_XT_OP(XtManage, XtManageChild)
DEFUN_XT_OP(XtUnmanage, XtUnmanageChild)
DEFUN_XT_OP(XtMap, XtMapWidget)
DEFUN_XT_OP(XtUnmap, XtUnmapWidget)
DEFUN_XT_OP(XtDestroy, XtDestroyWidget)
DEFUN_XT_OP(XtPopdown, XtPopdown)
DEFUN_XT_OP(XtUninstallTranslations, XtUninstallTranslations)

DEFUN_XT_2OP(XtAugmentTranslations, ANY, any)
DEFUN_XT_2OP(XtOverrideTranslations, ANY, any)
DEFUN_XT_2OP(XtPopup, INT, int)
DEFUN_XT_2OP(XtSetMappedWhenManaged, INT, int)
DEFUN_XT_2OP(XtSetSensitive, INT, int)

DEFUN(XtMove)
{
     GET_WIDGET_OR_LAYOUT("XtMove");
     arg2 = get_arg(aw, 2, INT, TRUE);
     if (value_type(arg2)) {
	  int x = value_int(arg2);

	  arg2 = get_arg(aw, 3, INT, TRUE);
	  if (value_type(arg2))
	       XtMoveWidget(tmp, x, value_int(arg2));
     }
     return(arg);
}

DEFUN(XtName)
{
     String str;

     GET_WIDGET_OR_LAYOUT("XtName");
     if ((str = XtName(tmp)) != NULL)
	  value_set(arg, STATIC, STRING, aobj, new_aobj(str));
     else
	  value_clear(arg);
     return(arg);
}

DEFUN(XtNameToWidget)
{
     Widget xret = (Widget)NULL;

     GET_WIDGET_OR_LAYOUT("XtNameToWidget");
     arg2 = get_arg(aw, 2, STRING, TRUE);
     if (value_type(arg2))
	  xret = XtNameToWidget(tmp, value_string(arg2));
     value_set(arg, DATA, WIDGET, widget, xret);
     return(arg);
}

DEFUN(XtResize)
{
     GET_WIDGET_OR_LAYOUT("XtResize");
     arg2 = get_arg(aw, 2, INT, TRUE);
     if (value_type(arg2)) {
	  int width = value_int(arg2);

	  arg2 = get_arg(aw, 3, INT, TRUE);
	  if (value_type(arg2)) {
	       int height = value_int(arg2);

	       arg2 = get_arg(aw, 4, INT, TRUE);
	       if (value_type(arg2))
		    XtResizeWidget(tmp, width, height, value_int(arg2));
	  }
     }
     return(arg);
}

DEFUN(XtSetSelectionTimeout)
{
     arg = get_arg(aw, 1, INT, FALSE);
     XtAppSetSelectionTimeout(XtWidgetToApplicationContext(aw),
			      value_int(arg));
     return(arg);
}

DEFUN(XtAddTimeOut)
{
     arg = get_arg(aw, 1, INT, FALSE);
     arg2 = get_arg(aw, 2, ANY, FALSE);
     if (value_type(arg2) != SYMBOL ||
	 symbol_class(value_symbol(arg2)) != FUNCTION)
	  exec_warn(aw, "Non-function passed to XtAddTimeOut");
     else {
	  Symbol sym;
	  Function f;

	  sym = value_symbol(arg2);
	  f = symbol_function(sym);
	  value_type(arg) = ANY;
	  switch(symbol_type(sym)) {
	  case FUNC_USER:
	       value_any(arg) =
		    (Generic)XtAppAddTimeOut(XtWidgetToApplicationContext(aw),
					     value_int(arg),
					     do_timer_user,
					     sym);
	       break;

	  case FUNC_BUILTIN:
	  case FUNC_EXTERN:
	       value_any(arg) =
		    (Generic)XtAppAddTimeOut(XtWidgetToApplicationContext(aw),
					     value_int(arg),
					     function_faddr(f),
					     function_data(f));
	       break;
	       
	  default:
	       exec_warn(aw, "Illegal function type for XtAddTimeOut");
	       break;
	  }
     }
     return(arg);
}

DEFUN(XtRemoveTimeOut)
{
     arg = get_arg(aw, 1, ANY, FALSE);
     XtRemoveTimeOut((XtIntervalId)value_any(arg));
     return(arg);
}

DEFUN(XtAddWorkProc)
{
     arg = get_arg(aw, 1, INT, FALSE);
     arg2 = get_arg(aw, 2, ANY, FALSE);
     if (value_type(arg2) != SYMBOL ||
	 symbol_class(value_symbol(arg2)) != FUNCTION)
	  exec_warn(aw, "Non-function passed to XtAddTimeOut");
     else {
	  Symbol sym;
	  Function f;

	  sym = value_symbol(arg2);
	  f = symbol_function(sym);
	  value_type(arg) = ANY;
	  switch(symbol_type(sym)) {
	  case FUNC_USER:
	       value_any(arg) =
		    (Generic)XtAppAddWorkProc(XtWidgetToApplicationContext(aw),
					      do_workproc_user,
					      sym);
	       break;

	  case FUNC_BUILTIN:
	  case FUNC_EXTERN:
	       value_any(arg) =
		    (Generic)XtAppAddWorkProc(XtWidgetToApplicationContext(aw),
					      function_faddr(f),
					      function_data(f));
	       break;
	       
	  default:
	       exec_warn(aw, "Illegal function type for XtAddWorkProc");
	       break;
	  }
     }
     return(arg);
}

DEFUN(XtRemoveWorkProc)
{
     arg = get_arg(aw, 1, ANY, FALSE);
     XtRemoveTimeOut((XtWorkProcId)value_any(arg));
     return(arg);
}

DEFUN(XtIsSubclass)
{
     GET_WIDGET_OR_LAYOUT("XtIsSubclass");
     arg = get_arg(aw, 2, ANY, FALSE);
     if (value_type(arg) == LAYOUT) {
	  value_type(arg) = INT;
	  value_int(arg) = XtIsSubclass(tmp,
					layout_wclass(aw, value_any(arg)));
     }
     else if (value_type(arg) == WCLASS) {
	  value_type(arg) = INT;
	  value_int(arg) = XtIsSubclass(tmp, value_wclass(arg));
     }
     else
	  exec_error(aw, "%s: Expected widget class or layout, got '%s'",
		     "XtIsSubclass", type_name(value_type(arg)));
     return(arg);
}
	  
Local Value wcreate(aw, manage)
AwlWidget aw;
Boolean manage;
{
     int i, nargs = awl_nparms(aw);
     String name, nm;
     Widget parent;
     WidgetClass class;
     int a;
     Arg args[50];

     if (manage)
	  nm = "XtCreateManaged";
     else
	  nm = "XtCreate";

     /* if argcnt = 1, it's a layout w/default parent */
     if (nargs == 1) {
	  arg = get_arg(aw, 1, LAYOUT, TRUE);
	  if (value_type(arg))
	       layout_create(aw, aw, value_any(arg), manage);
     }
     /* if argcnt == 2, it's a layout with the parent specified */
     else if (nargs == 2) {
	  arg = get_arg(aw, 1, WIDGET, TRUE);
	  arg2 = get_arg(aw, 2, LAYOUT, TRUE);
	  if (value_type(arg) && value_type(arg2))
	       layout_create(aw, value_widget(arg), value_any(arg2), manage);
     }
     /* otherwise it's a direct widget declaration */
     else {
	  if (nargs < 3)
	       exec_error(aw, "%s: Minimum of 3 args to create widget", nm);
	  else if (!(nargs % 2))
	       exec_error(aw, "%s: Incomplete argument pair for widget", nm);
	  else {
	       arg = get_arg(aw, 1, STRING, FALSE);
	       if (value_type(arg))
		    name = value_string(arg);
	       else
		    exec_error(aw, "%s: Missing or invalid widget name", nm);
	       arg = get_arg(aw, 2, WCLASS, TRUE);
	       class = value_wclass(arg);
	       if (!map_wclass_to_wname(class))
		    exec_error(aw, "%s: Invalid widget class specified", nm);
	       arg = get_arg(aw, 3, ANY, FALSE);
	       if (value_type(arg) == WIDGET)
		    parent = value_widget(arg);
	       else if (value_type(arg) == LAYOUT)
		    parent = layout_id(aw, value_any(arg));
	       else
		    exec_error(aw, "%s: Invalid parent specified", nm);
	       if (!parent)
		    exec_error(aw, "%s: parent widget is NULL", nm);
	       a = 0;
	       for (i = 4; i < nargs; i += 2) {
		    String rname;

		    arg = get_arg(aw, i, STRING, TRUE);
		    rname = value_string(arg);
		    wset_arg_by_value(args, &a, rname,
				      get_arg(aw, i + 1, ANY, FALSE));
	       }
	       value_set(arg, DATA, WIDGET, widget, 
			 ((*wcreate_rtn)(class, manage))(name, class,
							 parent, args, a));
	  }
     }
     return(arg);
}

/* Create a managed widget */
DEFUN(XtCreateManaged)
{
     return(wcreate(aw, TRUE));
}

/* Create an unmanaged widget */
DEFUN(XtCreate)
{
     return(wcreate(aw, FALSE));
}

/* Toot the horn */
DEFUN(XBell)
{
     arg = get_arg(aw, 1, INT, FALSE);
     XBell(XtDisplay(aw), value_int(arg));
     return(arg);
}
